/* Charts.java

	Purpose:
		
	Description:
		
	History:
		Wed, Jan 08, 2014  5:12:42 PM, Created by jumperchen

Copyright (C)  Potix Corporation. All Rights Reserved.

This program is distributed under LGPL Version 2.1 in the hope that
it will be useful, but WITHOUT ANY WARRANTY.
 */
package org.zkoss.chart;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import org.zkoss.chart.impl.PlotEngineImpl;
import org.zkoss.chart.model.CategoryModel;
import org.zkoss.chart.model.ChartsDataEvent;
import org.zkoss.chart.model.ChartsDataListener;
import org.zkoss.chart.model.ChartsModel;
import org.zkoss.chart.model.ChartsModelProxy;
import org.zkoss.chart.model.DialChartsDataEvent;
import org.zkoss.chart.model.DialChartsDataEventImpl;
import org.zkoss.chart.model.DialModel;
import org.zkoss.chart.model.FromToModel;
import org.zkoss.chart.model.SingleValueCategoryModel;
import org.zkoss.chart.model.WordCloudModel;
import org.zkoss.chart.model.XYModel;
import org.zkoss.chart.options3D.Options3D;
import org.zkoss.chart.plotOptions.PlotOptions;
import org.zkoss.chart.plotOptions.SeriesPlotOptions;
import org.zkoss.chart.util.JSFunction;
import org.zkoss.json.JSONValue;
import org.zkoss.json.JavaScriptValue;
import org.zkoss.lang.Library;
import org.zkoss.lang.Objects;
import org.zkoss.lang.Strings;
import org.zkoss.zk.au.AuRequest;
import org.zkoss.zk.au.AuRequests;
import org.zkoss.zk.ui.AbstractComponent;
import org.zkoss.zk.ui.UiException;
import org.zkoss.zk.ui.WebApps;
import org.zkoss.zk.ui.WrongValueException;
import org.zkoss.zk.ui.event.Event;
import org.zkoss.zk.ui.event.EventListener;
import org.zkoss.zk.ui.event.Events;
import org.zkoss.zk.ui.event.KeyEvent;
import org.zkoss.zk.ui.event.MouseEvent;
import org.zkoss.zul.ChartModel;
import org.zkoss.zul.event.ChartDataEvent;

/**
 * Charts is a ZK component that wraps the <a
 * href="http://www.highcharts.com/">Highcharts.com</a> charting library which
 * is written in pure HTML5/JavaScript, offering intuitive, interactive charts
 * to your web site or web application. Charts component currently supports
 * area, arearange, areaspline, areasplinerange, bar, bellcurve, boxplot, bubble,
 * bullet, column, columnpyramid, columnrange, cylinder, dependencywheel, errorbar,
 * funnel (funnel3d), gauge, heatmap, histogram, item, line, networkgraph, organization,
 * packedbubble, pareto, pie, polygon, pyramid (pyramid3d), sankey, scatter (scatter3d),
 * solidgauge, spline, streamgraph. sunburst, tilemap, timeline, treemap, variablepie,
 * variwide, vector, venn, windbarb, wordcloud, waterfall, and xrange chart types.
 * <p>
 * Model supports:
 * <ul>
 * <li>CategoryModel: line, area, bar, column, columnpyramid, areaspline, waterfall,
 * spline, streamgraph, bullet, cylinder</li>
 * <li>SingleValueCategoryModel: pie, pyramid (pyramid3d), funnel (funnel3d)</li>
 * <li>DialModel: gauge</li>
 * <li>XYModel: line, area, bar, column, columnpyramid, areaspline, scatter, bubble, waterfall,
 * spline, columnrange, errorbar, polygon, streamgraph, bullet, cylinder, windbarb</li>
 * <li>XYZModel: bubble, arearange, areasplinerange, heatmap, scatter3d, variwide</li>
 * <li>BoxPlotModel: boxplot</li>
 * <li>WordCloudModel: wordcloud</li>
 * <li>FromToModel: networkgraph, organization</li>
 * <li>FromToWeightModel: networkgraph, organization, sankey, dependencywheel</li>
 * </ul>
 * The model implementations use the same as Charts' API to manipulate the data
 * of the model, so developer can use the Charts' API instead of using model if
 * the chart data are not dynamically changed.
 * </p>
 * 
 * @author jumperchen
 * @author RaymondChao
 */
public class Charts extends AbstractComponent implements ChartsApi {

	private static final long serialVersionUID = 20140108153712L;
	// chart type
	/**
	 * Chart type: area
	 */
	public static final String AREA = "area";
	/**
	 * Chart type: area range
	 */
	public static final String AREA_RANGE = "arearange";
	/**
	 * Chart type: area spline
	 */
	public static final String AREA_SPLINE = "areaspline";
	/**
	 * Chart type: area spline range
	 */
	public static final String AREA_SPLINE_RANGE = "areasplinerange";
	/**
	 * Chart type: bar
	 */
	public static final String BAR = "bar";
	/**
	 * Chart type: bell curve
	 * @since 7.2.1.0
	 */
	public static final String BELLCURVE = "bellcurve";
	/**
	 * Chart type: boxplot
	 */
	public static final String BOX_PLOT = "boxplot";
	/**
	 * Chart type: bubble
	 */
	public static final String BUBBLE = "bubble";
	/**
	 * Chart type: column
	 */
	public static final String COLUMN = "column";
	/**
	 * Chart type: column pyramid
	 * @since 7.2.1.0
	 */
	public static final String COLUMN_PYRAMID = "columnpyramid";
	/**
	 * Chart type: column range
	 */
	public static final String COLUMN_RANGE = "columnrange";
	/**
	 * Chart type: cylinder
	 * @since 7.2.1.0
	 */
	public static final String CYLINDER = "cylinder";
	/**
	 * Chart type: dependency wheel
	 * @since 7.2.1.0
	 */
	public static final String DEPENDENCYWHEEL = "dependencywheel";
	/**
	 * Chart type: error bar
	 */
	public static final String ERROR_BAR = "errorbar";
	/**
	 * Chart type: funnel
	 */
	public static final String FUNNEL = "funnel";
	/**
	 * Chart type: funnel 3d
	 * @since 7.2.1.0
	 */
	public static final String FUNNEL3D = "funnel3d";
	/**
	 * Chart type: gauge
	 */
	public static final String GAUGE = "gauge";
	/**
	 * Chart type: histogram
	 * @since 7.2.1.0
	 */
	public static final String HISTOGRAM = "histogram";
	/**
	 * Chart type: item
	 * @since 7.2.1.0
	 */
	public static final String ITEM = "item";
	/**
	 * Chart type: solid gauge
	 */
	public static final String SOLID_GAUGE = "solidgauge";
	/**
	 * Chart type: line
	 */
	public static final String LINE = "line";
	/**
	 * Chart type: network graph
	 * @since 7.2.1.0
	 */
	public static final String NETWOKRGRAPH = "networkgraph";
	/**
	 * Chart type: organization
	 * @since 7.2.1.0
	 */
	public static final String ORGANIZATION = "organization";
	/**
	 * Chart type: pareto
	 * @since 7.2.1.0
	 */
	public static final String PARETO = "pareto";
	/**
	 * Chart type: pie
	 */
	public static final String PIE = "pie";
	/**
	 * Chart type: polygon
	 */
	public static final String POLYGON = "polygon";
	/**
	 * Chart type: sankey
	 * @since 7.2.1.0
	 */
	public static final String SANKEY = "sankey";
	/**
	 * Chart type: scatter
	 */
	public static final String SCATTER = "scatter";
	/**
	 * Chart type: scatter3d
	 */
	public static final String SCATTER3D = "scatter3d";
	/**
	 * Chart type: series
	 */
	public static final String SERIES = "series";
	/**
	 * Chart type: spline
	 */
	public static final String SPLINE = "spline";
	/**
	 * Chart type: stream graph
	 * @since 7.2.1.0
	 */
	public static final String STREAMGRAPH = "streamgraph";
	/**
	 * Chart type: sunburst
	 * @since 7.2.1.0
	 */
	public static final String SUNBURST = "sunburst";
	/**
	 * Chart type: tilemap
	 * @since 7.2.1.0
	 */
	public static final String TILEMAP = "tilemap";
	/**
	 * Chart type: timeline
	 * @since 7.2.1.0
	 */
	public static final String TIMELINE = "timeline";
	/**
	 * Chart type: waterfall
	 */
	public static final String WATERFALL = "waterfall";
	/**
	 * Chart type: packed bubble
	 * @since 7.2.1.0
	 */
	public static final String PACKEDBUBBLE = "packedbubble";
	/**
	 * Chart type: polar
	 *
	 * @deprecated since 3.0.2. Use {@link Chart#setPolar(boolean)} instead.
	 */
	@Deprecated
	public static final String POLAR = "polar";
	/**
	 * Chart type: pyramid
	 */
	public static final String PYRAMID = "pyramid";
	/**
	 * Chart type: pyramid3d
	 * @since 7.2.1.0
	 */
	public static final String PYRAMID3D = "pyramid3d";
	/**
	 * Chart type: heat map
	 */
	public static final String HEATMAP = "heatmap";
	/**
	 * Chart type: tree map
	 */
	public static final String TREEMAP = "treemap";
	/**
	 * Chart type: variable pie
	 * @since 7.2.1.0
	 */
	public static final String VARIABLEPIE = "variablepie";
	/**
	 * Chart type: variwide
	 * @since 7.2.1.0
	 */
	public static final String VARIWIDE = "variwide";
	/**
	 * Chart type: vector
	 * @since 7.2.1.0
	 */
	public static final String VECTOR = "vector";
	/**
	 * Chart type: venn
	 * @since 7.2.1.0
	 */
	public static final String VENN = "venn";
	/**
	 * Chart type: wind barb
	 * @since 7.2.1.0
	 */
	public static final String WINDBARB = "windbarb";
	/**
	 * Chart type: word cloud
	 * @since 7.2.1.0
	 */
	public static final String WORDCLOUD = "wordcloud";
	/**
	 * Chart type: xrange
	 * @since 7.2.1.0
	 */
	public static final String XRANGE = "xrange";
	
	/** The ZK CSS class. */
	protected String _zclass;

	static {
		// component events
		addClientEvent(Charts.class, Events.ON_CLICK, 0);
		addClientEvent(Charts.class, Events.ON_DOUBLE_CLICK, 0);
		addClientEvent(Charts.class, Events.ON_RIGHT_CLICK, 0);
		addClientEvent(Charts.class, Events.ON_MOUSE_OVER, 0); // not to use
																// CE_DUPLICATE_IGNORE
																// since
																// there is
																// an order
																// issue
		addClientEvent(Charts.class, Events.ON_MOUSE_OUT, 0);
		addClientEvent(Charts.class, Events.ON_AFTER_SIZE, CE_DUPLICATE_IGNORE);
		addClientEvent(Charts.class, Events.ON_OK, 0);
		addClientEvent(Charts.class, Events.ON_CANCEL, 0);
		addClientEvent(Charts.class, Events.ON_CTRL_KEY, 0);
		
		// chart events
		addClientEvent(Charts.class, Events.ON_SELECTION, 0);
		addClientEvent(Charts.class, ChartsEvents.ON_PLOT_CLICK, 0);
		addClientEvent(Charts.class, ChartsEvents.ON_PLOT_CHECKBOX_CLICK, 0);
		addClientEvent(Charts.class, ChartsEvents.ON_PLOT_HIDE, CE_IMPORTANT);
		addClientEvent(Charts.class, ChartsEvents.ON_PLOT_LEGEND_ITEM_CLICK, 0);
		addClientEvent(Charts.class, ChartsEvents.ON_PLOT_MOUSE_OUT, 0);
		addClientEvent(Charts.class, ChartsEvents.ON_PLOT_MOUSE_OVER, 0);
		addClientEvent(Charts.class, ChartsEvents.ON_PLOT_SHOW, CE_IMPORTANT);
		addClientEvent(Charts.class, ChartsEvents.ON_PLOT_DRILL_DOWN, 0);
		addClientEvent(Charts.class, ChartsEvents.ON_PLOT_DRILL_UP, 0);
		addClientEvent(Charts.class, ChartsEvents.ON_PLOT_MOUSE_OVER_POINT, 0);
		addClientEvent(Charts.class, ChartsEvents.ON_PLOT_MOUSE_OUT_POINT, 0);
		
		// Update point select information
		addClientEvent(Charts.class, ChartsEvents.ON_PLOT_SELECT, CE_IMPORTANT);
		addClientEvent(Charts.class, ChartsEvents.ON_PLOT_UNSELECT, CE_IMPORTANT);
	}

	private boolean _smartDrawChart; // whether post the smartDraw event
										// already?
	private ChartsDataListener _dataListener;

	// chart data model
	private ChartsModel _model; // chart data model

	private PlotData _plotData;
	private Options _options;
	private Theme _theme;
	private ChartsApi _chartsApi;
	
	private String _hflex;
	private String _vflex;
	private String _ctrlKeys;
	
	
	/** A list of selected points. */
	private transient Set<Point> _selPoints;

	private static final String ZKCHARTS_LOCK = "org.zkoss.chart.rt.lock.key";
	static { Library.setProperty(ZKCHARTS_LOCK, "true"); }
	private static final Random _RGEN = new Random(new java.util.Date().getTime());
	/**package*/ boolean checkLock() {
		return _RGEN.nextInt() > 0 || "true".equals(Library.getProperty(ZKCHARTS_LOCK));
	}
	
	public Charts() {
		_selPoints = new HashSet<Point>();
		_chartsApi = new ChartsEngine(this);
	}

	/**
	 * Returns the plot data which is a JSON object to collect all chart's
	 * options for Charts JS.
	 */
	public PlotData getPlotData() {
		if (_plotData == null)
			_plotData = _chartsApi.getPlotData();
		return _plotData;
	}

	/**
	 * Returns the drilldown object.
	 */
	public Drilldown getDrilldown() {
		return _chartsApi.getDrilldown();
	}

	/**
	 * Sets the global options for all Charts in the same browser page.
	 */
	public void setOptions(Options options) {
		if (options == null)
			throw new WrongValueException("Global options cannot be null!");

		_options = options; // no need to check, force to update
		invalidate();
	}

	/**
	 * Sets the formats a JavaScript date timestamp (milliseconds since Jan 1st
	 * 1970) into a human readable date string. The format is a subset of the
	 * formats for PHP's <a
	 * href="http://php.net/manual/en/function.strftime.php">strftime</a>
	 * function.
	 * 
	 * @param format
	 *            A string containing some of the formats above.
	 * @param time
	 *            The JavaScript time to format.
	 * @param capitalize
	 *            Whether to capitalize words in the return string.
	 */
	public void setDateFormat(String format, Number time, boolean capitalize) {
		addBeforeInvokeScript("Highcharts.dateFormat('"
				+ JSONValue.toJSONString(format) + "', "
				+ JSONValue.toJSONString(time) + ", "
				+ JSONValue.toJSONString(capitalize) + ")");
		invalidate();
	}

	/**
	 * Formats a JavaScript number with grouped thousands, a fixed amount of
	 * decimals and an optional decimal point. It is a port of PHP's function
	 * with the same name. See PHP number_format for a full explanation of the
	 * parameters.
	 * 
	 * @param number
	 *            The raw number to format.
	 * @param decimals
	 *            The desired number of decimals.
	 * @param decimalPoint
	 *            The decimal point. Defaults to "." or to the string specified
	 *            globally in {@link Lang#setDecimalPoint(String)}.
	 * @param thousandsSep
	 *            The thousands separator. Defaults to " " or to the string
	 *            specified globally in {@link Lang#setThousandsSep(String)}
	 */
	public void setNumberFormat(Number number, Number decimals,
			String decimalPoint, String thousandsSep) {
		addBeforeInvokeScript(String.format(
				"Highcharts.numberFormat(%1$s,%2$s,%3$s,%4$s)",
				JSONValue.toJSONString(number),
				JSONValue.toJSONString(decimals),
				JSONValue.toJSONString(decimalPoint),
				JSONValue.toJSONString(thousandsSep)));

		invalidate();
	}

	private List<JavaScriptValue> _befInvoke = new LinkedList<JavaScriptValue>();

	private void addBeforeInvokeScript(String script) {
		JSFunction jsFunction = new JSFunction();
		jsFunction.evalJavascript(script);
		_befInvoke.add(new JavaScriptValue(jsFunction.toExecFunction()));
	}

	/**
	 * Sets the default colors for the chart's series. When all colors are used,
	 * new colors are pulled from the start again.
	 */
	public void setColors(List<Color> colors) {
		_chartsApi.setColors(colors);
	}

	/**
	 * Sets the default colors for the chart's series. When all colors are used,
	 * new colors are pulled from the start again.
	 */
	public void setColors(String... source) {
		_chartsApi.setColors(source);
	}

	/**
	 * Returns the default colors for the chart's series. When all colors are
	 * used, new colors are pulled from the start again.
	 */
	public List<Color> getColors() {
		return _chartsApi.getColors();
	}

	/**
	 * Sets the theme options for all Charts in the same browser page.
	 */
	public void setTheme(Theme theme) {
		_chartsApi.setTheme(theme);
		_theme = theme; // no need to check, force to update
		Events.postEvent(ChartsEvents.ON_PLOT_THEME_CHANGE, this, null);
		invalidate();
	}

	/**
	 * Returns the global theme for all Charts.
	 * <p>
	 * Default is null
	 */
	public Theme getTheme() {
		return _theme;
	}

	/**
	 * Returns the global options for all Charts.
	 * <p>
	 * Default is null.
	 */
	public Options getOptions() {
		return _options;
	}

	/**
	 * Returns the chart model associated with this chart, or null if this chart
	 * is not associated with any chart data model.
	 */
	public ChartsModel getModel() {
		return _model;
	}

	/**
	 * Sets the chart model associated with this chart. If a non-null model is
	 * assigned, no matter whether it is the same as the previous, it will
	 * always cause re-render.
	 * 
	 * @param model
	 *            the chart model to associate, or null to dissociate any
	 *            previous model.
	 * @exception UiException
	 *                if failed to initialize with the model
	 */
	public void setModel(ChartsModel model) {
		if (_model != model) {
			if (_model != null) {
				_model.removeChartsDataListener(_dataListener);
			}
			invalidate(); // reset all chart data
			_model = model;
			initDataListener();
			smartDrawChart();
		}
	}

	public void setModel(ChartModel model) {
		if (model != null) {
			final String version = WebApps.getCurrent().getVersion().replace(".", "");
			int f = Integer.parseInt(String.valueOf(version.charAt(0)));
			if (f < 7) {
				throw new IllegalArgumentException(
						"The version of ChartModel is not supported, please use ZK 7.0.1+ version or new ChartsModel instead. ");
			} else if (f == 7) {
				int s = Integer.parseInt(String.valueOf(version.charAt(1)));
				int t = Integer.parseInt(String.valueOf(version.charAt(2)));
				if (s == 0 && t == 0)
					throw new IllegalArgumentException(
							"The version of ChartModel is not supported, please use ZK 7.0.1+ version or new ChartsModel instead. ");
			}
			setModel(ChartsModelProxy.getInstance(model));
		} else {
			setModel((ChartsModel) null);
		}
	}

	private void initDataListener() {
		if (_model != null) {
			if (_dataListener == null) {
				_dataListener = new MyChartsDataListener();
			}
			_model.addChartsDataListener(_dataListener);
		}
	}
	private static final String ATTR_ON_SYNC_MODEL_LATER = "org.zkoss.chart.onSyncModelLater";
	private void postonSyncModelLater() {
		if (getAttribute(ATTR_ON_SYNC_MODEL_LATER) == null) {
			setAttribute(ATTR_ON_SYNC_MODEL_LATER, Boolean.TRUE);
			Events.postEvent("onSyncModel", this, null);
		}
	}
	public void onSyncModel() {
		removeAttribute(ATTR_ON_SYNC_MODEL_LATER);
		if (_model != null) {
			int pSize = 0;
			boolean rerender = false;
			for (int size = getPlotData().getSeriesSize(), i = 0; i < size; i++) {
				if (pSize == 0) {
					List<Point> data = getSeries(i).getData();
					pSize = data != null ? data.size() : 0;
				} else { // the series size is not matched , we need to re-render the whole model again for Bug ZKCHARTS-10
					List<Point> data = getSeries(i).getData();
					if (data != null && pSize != data.size()) {
						rerender = true;
						break;
					}
				}
			}
			if (rerender) {
				smartDrawChart();
			}
		}
	}

	private class MyChartsDataListener implements ChartsDataListener,
			Serializable {
		private static final long serialVersionUID = 20091008183622L;

		public void onChange(ChartsDataEvent event) {

			ChartsModel model = getModel();
			if (model instanceof SingleValueCategoryModel
					|| model instanceof XYModel
					|| model instanceof CategoryModel
					|| model instanceof FromToModel) {
				switch (event.getType()) {
				case ChartDataEvent.ADDED:
					int sIndex = event.getSeriesIndex();
					if (event.getCategory() instanceof Number) {
						if (event.getData() instanceof Number[]) {
							Number[] numbers = (Number[]) event.getData();
							if (numbers.length == 3)
								getSeries(sIndex).addPoint(numbers[0],
										numbers[1], numbers[2]);
							else if (numbers.length == 2)
								getSeries(sIndex).addPoint(numbers[0],
										numbers[1]);
						} else {
							getSeries(sIndex).addPoint(
									(Number) event.getCategory(),
									(Number) event.getData());
						}
					} else {
						if (model instanceof CategoryModel) {
							for (int i = 0, len = getSeriesSize(); i < len; i++) {
								getSeries(i).addPoint(
										i == sIndex ? (Number) event.getData() : null);
							}
						} else if (model instanceof WordCloudModel) {
							Point p = new Point();
							p.setName(event.getCategory().toString());
							p.setWeight((Number) event.getData());
							getSeries(sIndex).addPoint(p);
						} else if (model instanceof FromToModel) {
							Point p = new Point();
							p.setFrom(event.getCategory().toString());
							if (event.getData() instanceof Object[]) {
								Object[] data = (Object[]) event.getData();
								p.setTo((String) data[0]);
								p.setWeight((Number) data[1]);
							} else {
								p.setTo((String) event.getData());
							}
							getSeries(sIndex).addPoint(p);
						} else {
							for (int i = 0, len = getSeriesSize(); i < len; i++) {
								getSeries(i).addPoint(
										(String) event.getCategory(),
										i == sIndex ? (Number) event.getData() : null);
							}
						}

						if (getXAxis().getCategories() != null) {
							List<String> categories = getXAxis().getCategories();
							if (!categories.contains(event.getCategory())) {
								List<String> newList = new ArrayList<String>(categories);
								newList.add((String) event.getCategory());
								getXAxis().setCategories(newList);
							}
						}
					}
					if (event.getSeries() != null)
						getSeries(sIndex).setName(event.getSeries().toString());
					postonSyncModelLater();
					break;
				case ChartDataEvent.CHANGED:
					int cIndex = event.getCategoryIndex();
					if (event.getData() instanceof Number[]) {
						Number[] numbers = (Number[]) event.getData();
						getSeries(event.getSeriesIndex()).getPoint(cIndex)
								.update(numbers[0], numbers[1], numbers[2]);
					} else if (model instanceof XYModel) {
						getSeries(event.getSeriesIndex()).getPoint(cIndex)
								.update((Number) event.getCategory(),
										(Number) event.getData());
					} else if (model instanceof WordCloudModel) {
						getSeries(event.getSeriesIndex()).getPoint(cIndex)
								.update((String) event.getCategory(),
										(Number) event.getData());
					} else if (model instanceof FromToModel) {
						Point p = getSeries(event.getSeriesIndex()).getPoint(cIndex);
						if (event.getData() instanceof Object[]) {
							Object[] data = (Object[]) event.getData();
							p.update((String) event.getCategory(), (String) data[0], (Number) data[1]);
						} else {
							p.update((String) event.getCategory(), (String) event.getData());
						}
					} else {
						getSeries(event.getSeriesIndex()).getPoint(cIndex)
								.update((Number) event.getData());
					}
					postonSyncModelLater();
					break;
				case ChartDataEvent.REMOVED:
					int cIndex2 = event.getCategoryIndex();
					int sIndex2 = event.getSeriesIndex();
					if (sIndex2 < 0) {
						for (int i = getSeriesSize(); --i >= 0;)
							getSeries(i).remove();

						// reset all
						List<String> categories = getXAxis().getCategories();
						if (categories != null) {
							getXAxis().setCategories(new ArrayList<String>()); // force to update
						}
					} else {
						Series series2 = getSeries(sIndex2);
						// xyz model
						if (cIndex2 < 0) {
							series2.remove();
						} else {
							series2.getPoint(cIndex2).remove();
							if (series2.getData().isEmpty())
								series2.remove();
						}
					}
					postonSyncModelLater();
					break;
				}
			} else if (model instanceof DialModel) {
				if (event instanceof DialChartsDataEvent) {
					DialChartsDataEventImpl evt = (DialChartsDataEventImpl) event;
					String key = evt.getPropertyKey();
					if (DialChartsDataEvent.SCALE_VALUE.equals(key)) {
						getSeries().getPoint(0).update((Number) evt.getData());
					} else if (DialChartsDataEvent.SCALE_TEXT.equals(key)) {
						getYAxis().getTitle().setText((String) evt.getData());
					} else if (DialChartsDataEvent.NEEDLE_COLOR.equals(key)) {
						getPlotOptions().getGauge().getDial()
								.setBackgroundColor((String) evt.getData());
					} else {
						smartDrawChart();
						invalidate();
					}
				}
			} else {
				throw new UiException("Unsupported with the model type ["
						+ model + "]");
			}
		}
	}

	/**
	 * Shows the loading dim the chart'ts plot area with the given text after
	 * rendering time, if any.
	 */
	public void showLoading(String text) {
		JSFunction js = new JSFunction();
		js.callFunction("showLoading", text);
		smartUpdate("eval", new JavaScriptValue(js.toWrapFunction()), true);
	}

	/**
	 * Hides the loading dim the chart'ts plot area after rendering time.
	 */
	public void hideLoading() {
		JSFunction js = new JSFunction();
		js.callFunction("hideLoading");
		smartUpdate("eval", new JavaScriptValue(js.toWrapFunction()), true);
	}

	/**
	 * Clears away other elements in the page and prints the chart as it is
	 * displayed.
	 */
	public void print() {
		JSFunction js = new JSFunction();
		js.callFunction("print");
		smartUpdate("eval", new JavaScriptValue(js.toWrapFunction()), true);
	}

	/**
	 * Set the chart's type (Charts.PIE, Charts.BAR, Charts.LINE, etc.).
	 * <p>
	 * Default: line.
	 */
	public void setType(String type) {
		Chart chart = getChart();
		String oldType = chart.getType();
		if (!Objects.equals(oldType, type)) {
			// Highcharts JS didn't have a type for Polar, so we need to do some
			// tricks.
			if (POLAR.equals(type))
				chart.setPolar(true); // just in case
			if (POLAR.equals(oldType))
				chart.setPolar(false);

			chart.setType(type);
			_chartImpl = null; // reset if changed
			smartDrawChart();
		}
	}

	/**
	 * Get the chart's type.
	 * <p>
	 * Default: line.
	 */
	public String getType() {
		return _chartsApi.getType();
	}

	/**
	 * Sets when using multiple axis, the ticks of two or more opposite axes
	 * will automatically be aligned by adding ticks to the axis or axes with
	 * the least ticks. This can be prevented by setting <code>alignTicks</code>
	 * to false. If the grid lines look messy, it's a good idea to hide them for
	 * the secondary axis by setting <code>gridLineWidth</code> to 0.
	 * 
	 * @param alignTicks
	 *            when using multiple axis, the ticks of two or more opposite
	 *            axes will automatically be aligned by adding ticks to the axis
	 *            or axes with the least ticks
	 * @see Chart#setAlignTicks(boolean)
	 */
	public void setAlignTicks(boolean alignTicks) {
		_chartsApi.setAlignTicks(alignTicks);
	}

	/**
	 * Returns when using multiple axis, the ticks of two or more opposite axes
	 * will automatically be aligned by adding ticks to the axis or axes with
	 * the least ticks. This can be prevented by setting <code>alignTicks</code>
	 * to false. If the grid lines look messy, it's a good idea to hide them for
	 * the secondary axis by setting <code>gridLineWidth</code> to 0.
	 * <p>
	 * Default: true.
	 * 
	 * @see Chart#isAlignTicks()
	 */
	public boolean isAlignTicks() {
		return _chartsApi.isAlignTicks();
	}

	/**
	 * Animation can be disabled throughout the chart by setting it to false
	 * here. It can be overridden for each individual API method as a function
	 * parameter. The only animation not affected by this option is the initial
	 * series animation, see {@link SeriesPlotOptions#setAnimation(boolean)}.
	 * 
	 * @see Chart#setAnimation(boolean)
	 * @see Animation
	 */
	public void setAnimation(boolean animation) {
		_chartsApi.setAnimation(animation);
	}

	/**
	 * Returns whether enable the animation
	 * <p>
	 * Default: true
	 */
	public boolean isAnimation() {
		return _chartsApi.isAnimation();
	}

	/**
	 * Sets to ignore the hidden series, if true, the axes will scale to the
	 * remaining visible series once one series is hidden. If false, hiding and
	 * showing a series will not affect the axes or the other series. For
	 * stacks, once one series within the stack is hidden, the rest of the stack
	 * will close in around it even if the axis is not affected.
	 * 
	 * @param ignoreHiddenSeries
	 *            if true, the axes will scale to the remaining visible series
	 *            once one series is hidden
	 */
	public void setIgnoreHiddenSeries(boolean ignoreHiddenSeries) {
		_chartsApi.setIgnoreHiddenSeries(ignoreHiddenSeries);
	}

	/**
	 * Returns if true, the axes will scale to the remaining visible series once
	 * one series is hidden. If false, hiding and showing a series will not
	 * affect the axes or the other series. For stacks, once one series within
	 * the stack is hidden, the rest of the stack will close in around it even
	 * if the axis is not affected.
	 * <p>
	 * Default: true.
	 */
	public boolean isIgnoreHiddenSeries() {
		return _chartsApi.isIgnoreHiddenSeries();
	}

	/**
	 * Sets the background color for the outer chart area.
	 * 
	 * @param backgroundColor
	 *            the background color for the outer chart area
	 */
	public void setBackgroundColor(String backgroundColor) {
		_chartsApi.setBackgroundColor(backgroundColor);
	}

	/**
	 * Sets the background gradient for the outer chart area.
	 */
	public void setBackgroundColor(LinearGradient gradient) {
		_chartsApi.setBackgroundColor(gradient);
	}

	/**
	 * Sets the background gradient for the outer chart area.
	 */
	public void setBackgroundColor(RadialGradient gradient) {
		_chartsApi.setBackgroundColor(gradient);
	}

	/**
	 * Sets the background gradient for the outer chart area.
	 */
	public void setBackgroundColor(Color color) {
		_chartsApi.setBackgroundColor(color);
	}

	/**
	 * Returns the background color for the outer chart area.
	 * <p>
	 * Default: "#FFFFFF".
	 * 
	 * @see Chart#getBackgroundColor()
	 */
	public Color getBackgroundColor() {
		return _chartsApi.getBackgroundColor();
	}

	/**
	 * Sets the color of the outer chart border. The border is painted using
	 * vector graphic techniques to allow rounded corners.
	 */
	public void setBorderColor(String borderColor) {
		_chartsApi.setBorderColor(borderColor);
	}

	/**
	 * Sets the color of the outer chart border. The border is painted using
	 * vector graphic techniques to allow rounded corners.
	 */
	public void setBorderColor(LinearGradient borderColor) {
		_chartsApi.setBorderColor(borderColor);
	}

	/**
	 * Sets the color of the outer chart border. The border is painted using
	 * vector graphic techniques to allow rounded corners.
	 */
	public void setBorderColor(RadialGradient borderColor) {
		_chartsApi.setBorderColor(borderColor);
	}

	/**
	 * Sets the color of the outer chart border. The border is painted using
	 * vector graphic techniques to allow rounded corners.
	 */
	public void setBorderColor(Color borderColor) {
		_chartsApi.setBorderColor(borderColor);
	}

	/**
	 * Returns the color of the outer chart border. The border is painted using
	 * vector graphic techniques to allow rounded corners.
	 * <p>
	 * Default: "#4572A7".
	 * 
	 * @see Chart#getBorderColor()
	 */
	public Color getBorderColor() {
		return _chartsApi.getBorderColor();
	}

	/**
	 * Sets the corner radius of the outer chart border. In export, the radius
	 * defaults to 0.
	 * <p>
	 * Default: 5.
	 * 
	 * @param borderRadius
	 *            the corner radius of the outer chart border
	 * @see Chart#setBorderRadius(Number)
	 */
	public void setBorderRadius(Number borderRadius) {
		_chartsApi.setBorderRadius(borderRadius);
	}

	/**
	 * Returns the corner radius of the outer chart border. In export, the
	 * radius defaults to 0.
	 * <p>
	 * Default: 5.
	 * 
	 * @see Chart#getBorderRadius()
	 */
	public Number getBorderRadius() {
		return _chartsApi.getBorderRadius();
	}

	/**
	 * Sets the pixel width of the outer chart border. The border is painted
	 * using vector graphic techniques to allow rounded corners.
	 * 
	 * @see Chart#setBorderWidth(Number)
	 */
	public void setBorderWidth(Number borderWidth) {
		_chartsApi.setBorderWidth(borderWidth);
	}

	/**
	 * Returns the pixel width of the outer chart border. The border is painted
	 * using vector graphic techniques to allow rounded corners.
	 * <p>
	 * Default: 0.
	 * 
	 * @see Chart#getBorderWidth()
	 */
	public Number getBorderWidth() {
		return _chartsApi.getBorderWidth();
	}

	/**
	 * Sets an explicit width for the chart. By default the width is calculated
	 * from the offset width of the containing element.
	 * 
	 * @param width
	 *            an explicit width for the chart
	 * @see Chart#setWidth(Number)
	 */
	public void setWidth(Number width) {
		_chartsApi.setWidth(width);
	}

	/**
	 * Returns an explicit width for the chart. By default the width is
	 * calculated from the offset width of the containing element.
	 * <p>
	 * Default: null.
	 * 
	 * @see Chart#getWidth()
	 */
	public Number getWidth() {
		return _chartsApi.getWidth();
	}

	/**
	 * Sets an explicit height for the chart. By default the height is
	 * calculated from the offset height of the containing element, or 400
	 * pixels if the containing element's height is 0.
	 * 
	 * @param height
	 *            an explicit height for the chart
	 * @see Chart#setHeight(Number)
	 */
	public void setHeight(Number height) {
		_chartsApi.setHeight(height);
	}

	public void setHeight(String height) {
		_chartsApi.setHeight(height);
	}

	/**
	 * Returns an explicit height for the chart. By default the height is
	 * calculated from the offset height of the containing element, or 400
	 * pixels if the containing element's height is 0.
	 * <p>
	 * Default: null.
	 * 
	 * @see Chart#getHeight()
	 */
	public Number getHeight() {
		return _chartsApi.getHeight();
	}

	/**
	 * Sets whether to invert the axes so that the x axis is vertical and y axis
	 * is horizontal. When true, the x axis is reversed by default. If a bar
	 * series is present in the chart, it will be inverted automatically.
	 * 
	 * @see Chart#setInverted(boolean)
	 */
	public void setInverted(boolean inverted) {
		_chartsApi.setInverted(inverted);
	}

	/**
	 * Returns whether to invert the axes so that the x axis is vertical and y
	 * axis is horizontal. When true, the x axis is reversed by default. If a
	 * bar series is present in the chart, it will be inverted automatically.
	 * <p>
	 * Default: false.
	 * 
	 * @see Chart#isInverted()
	 */
	public boolean isInverted() {
		return _chartsApi.isInverted();
	}

	/**
	 * Sets the margin between the outer edge of the chart and the plot area.
	 * The numbers in the array designate top, right, bottom and left
	 * respectively. Use the methods of {@link #setMarginTop(Number)},
	 * {@link #setMarginRight(Number)}, {@link #setMarginBottom(Number)} and
	 * {@link #setMarginLeft(Number)} for shorthand setting of one option.
	 * <p>
	 * Default: null.
	 * 
	 * @see #setMargin(Number[])
	 */
	public void setMargin(String margins) {
		_chartsApi.setMargin(margins);
	}
	
	/**
	 * Sets the margin between the outer edge of the chart and the plot area.
	 * 
	 */
	public void setMargin(Number margin) {
		_chartsApi.setMargin(margin);
	}

	/**
	 * Sets the margin between the outer edge of the chart and the plot area.
	 * The numbers in the array designate top, right, bottom and left
	 * respectively. Use the methods of {@link #setMarginTop(Number)},
	 * {@link #setMarginRight(Number)}, {@link #setMarginBottom(Number)} and
	 * {@link #setMarginLeft(Number)} for shorthand setting of one option.
	 * <p>
	 * Default: null.
	 * 
	 * @see #setMargin(String)
	 * @see Chart#setMargin(Number[])
	 */
	public void setMargin(Number[] margin) {
		_chartsApi.setMargin(margin);
	}

	/**
	 * Returns the margin between the outer edge of the chart and the plot area.
	 * The numbers in the array designate top, right, bottom and left
	 * respectively. Use the methods of {@link #setMarginTop(Number)},
	 * {@link #setMarginRight(Number)}, {@link #setMarginBottom(Number)} and
	 * {@link #setMarginLeft(Number)} for shorthand setting of one option.
	 * <p>
	 * Default: null.
	 * 
	 * @see Chart#getMargin()
	 */
	public Number[] getMargin() {
		return _chartsApi.getMargin();
	}

	/**
	 * Sets the margin between the bottom outer edge of the chart and the plot
	 * area. Use this to set a fixed pixel value for the margin as opposed to
	 * the default dynamic margin. See also {@link #setSpacingBottom(Number)}.
	 * 
	 * @param marginBottom
	 *            the margin between the bottom outer edge of the chart and the
	 *            plot area
	 * @see Chart#setMarginBottom(Number)
	 */
	public void setMarginBottom(Number marginBottom) {
		_chartsApi.setMarginBottom(marginBottom);
	}

	/**
	 * Returns the margin between the bottom outer edge of the chart and the
	 * plot area.
	 * <p>
	 * Default: null.
	 * 
	 * @see Chart#getMarginBottom()
	 */
	public Number getMarginBottom() {
		return _chartsApi.getMarginBottom();
	}

	/**
	 * Sets the margin between the left outer edge of the chart and the plot
	 * area. Use this to set a fixed pixel value for the margin as opposed to
	 * the default dynamic margin. See also {@link #setSpacingLeft(Number)}
	 * 
	 * @param marginLeft
	 *            the margin between the left outer edge of the chart and the
	 *            plot area
	 * @see Chart#setMarginLeft(Number)
	 */
	public void setMarginLeft(Number marginLeft) {
		_chartsApi.setMarginLeft(marginLeft);
	}

	/**
	 * Returns the margin between the left outer edge of the chart and the plot
	 * area.
	 * <p>
	 * Default: null.
	 * 
	 * @see Chart#getMarginLeft()
	 */
	public Number getMarginLeft() {
		return _chartsApi.getMarginLeft();
	}

	/**
	 * Sets the margin between the right outer edge of the chart and the plot
	 * area. Use this to set a fixed pixel value for the margin as opposed to
	 * the default dynamic margin. See also {@link #setSpacingRight(Number)}.
	 * 
	 * @param marginRight
	 *            the margin between the right outer edge of the chart and the
	 *            plot area
	 * @see Chart#setMarginRight(Number)
	 */
	public void setMarginRight(Number marginRight) {
		_chartsApi.setMarginRight(marginRight);
	}

	/**
	 * Returns the margin between the right outer edge of the chart and the plot
	 * area.
	 * <p>
	 * Default: null.
	 * 
	 * @see Chart#getMarginRight()
	 */
	public Number getMarginRight() {
		return _chartsApi.getMarginRight();
	}

	/**
	 * Sets the margin between the top outer edge of the chart and the plot
	 * area. Use this to set a fixed pixel value for the margin as opposed to
	 * the default dynamic margin. See also {@link #setSpacingTop(Number)}.
	 * 
	 * @param marginTop
	 *            the margin between the top outer edge of the chart and the
	 *            plot area
	 * @see Chart#setMarginTop(Number)
	 */
	public void setMarginTop(Number marginTop) {
		_chartsApi.setMarginTop(marginTop);
	}

	/**
	 * Returns the margin between the top outer edge of the chart and the plot
	 * area.
	 * <p>
	 * Default: null.
	 * 
	 * @see Chart#getMarginTop()
	 */
	public Number getMarginTop() {
		return _chartsApi.getMarginTop();
	}

	/**
	 * Sets equivalent to {@link #setZoomType(String)}, but for multitouch
	 * gestures only. By default, the <code>pinchType</code> is the same as the
	 * <code>zoomType</code> setting. However, pinching can be enabled
	 * separately in some cases, for example in stock charts where a mouse drag
	 * pans the chart, while pinching is enabled.
	 * <p>
	 * Default: null.
	 * 
	 * @see #setZoomType(String)
	 * @see Chart#setPinchType(String)
	 */
	public void setPinchType(String pinchType) {
		_chartsApi.setPinchType(pinchType);
	}

	/**
	 * Returns equivalent to {@link #getZoomType()}, but for multitouch gestures
	 * only. By default, the <code>pinchType</code> is the same as the
	 * <code>zoomType</code> setting. However, pinching can be enabled
	 * separately in some cases, for example in stock charts where a mouse drag
	 * pans the chart, while pinching is enabled.
	 * <p>
	 * Default: null.
	 * 
	 * @see Chart#getPinchType()
	 */
	public String getPinchType() {
		return _chartsApi.getPinchType();
	}

	/**
	 * Sets the background color for the plot area.
	 */
	public void setPlotBackgroundColor(String plotBackgroundColor) {
		_chartsApi.setPlotBackgroundColor(plotBackgroundColor);
	}

	/**
	 * Sets the background color for the plot area.
	 */
	public void setPlotBackgroundColor(LinearGradient plotBackgroundColor) {
		_chartsApi.setPlotBackgroundColor(plotBackgroundColor);
	}

	/**
	 * Sets the background color for the plot area.
	 */
	public void setPlotBackgroundColor(RadialGradient plotBackgroundColor) {
		_chartsApi.setPlotBackgroundColor(plotBackgroundColor);
	}

	/**
	 * Sets the background color for the plot area.
	 */
	public void setPlotBackgroundColor(Color plotBackgroundColor) {
		_chartsApi.setPlotBackgroundColor(plotBackgroundColor);
	}

	/**
	 * Returns the background color for the plot area.
	 * <p>
	 * Default: null.
	 * 
	 * @see Chart#getPlotBackgroundColor()
	 */
	public Color getPlotBackgroundColor() {
		return _chartsApi.getPlotBackgroundColor();
	}

	/**
	 * Sets the URL for an image to use as the plot background. To set an image
	 * as the background for the entire chart, set a CSS background image to the
	 * container element. Note that for the image to be applied to exported
	 * charts, its URL needs to be accessible by the export server.
	 * 
	 * @param plotBackgroundImage
	 *            the URL for an image to use as the plot background
	 * @see Chart#setPlotBackgroundImage(String)
	 */
	public void setPlotBackgroundImage(String plotBackgroundImage) {
		_chartsApi.setPlotBackgroundImage(plotBackgroundImage);
	}

	/**
	 * Returns the URL for an image to use as the plot background.
	 * <p>
	 * Default: null.
	 * 
	 * @see Chart#getPlotBackgroundImage()
	 */
	public String getPlotBackgroundImage() {
		return _chartsApi.getPlotBackgroundImage();
	}

	/**
	 * Sets the color of the inner chart or plot area border.
	 */
	public void setPlotBorderColor(String plotBorderColor) {
		_chartsApi.setPlotBorderColor(plotBorderColor);
	}

	/**
	 * Sets the color of the inner chart or plot area border.
	 */
	public void setPlotBorderColor(LinearGradient plotBorderColor) {
		_chartsApi.setPlotBorderColor(plotBorderColor);
	}

	/**
	 * Sets the color of the inner chart or plot area border.
	 */
	public void setPlotBorderColor(RadialGradient plotBorderColor) {
		_chartsApi.setPlotBorderColor(plotBorderColor);
	}

	/**
	 * Sets the color of the inner chart or plot area border.
	 */
	public void setPlotBorderColor(Color plotBorderColor) {
		_chartsApi.setPlotBorderColor(plotBorderColor);
	}

	/**
	 * Returns the color of the inner chart or plot area border.
	 * <p>
	 * Default: "#C0C0C0".
	 * 
	 * @see Chart#getPlotBorderColor()
	 */
	public Color getPlotBorderColor() {
		return _chartsApi.getPlotBorderColor();
	}

	/**
	 * Sets the pixel width of the plot area border.
	 * 
	 * @param plotBorderWidth
	 *            the pixel width of the plot area border
	 * @see Chart#setPlotBorderWidth(Number)
	 */
	public void setPlotBorderWidth(Number plotBorderWidth) {
		_chartsApi.setPlotBorderWidth(plotBorderWidth);
	}

	/**
	 * Returns the pixel width of the plot area border.
	 * <p>
	 * Default: 0.
	 * 
	 * @see Chart#getPlotBorderWidth()
	 */
	public Number getPlotBorderWidth() {
		return _chartsApi.getPlotBorderWidth();
	}

	/**
	 * Sets whether to apply a drop shadow to the plot area. Requires that
	 * {@link #setPlotBackgroundColor(String)} be set.
	 * 
	 * @param plotShadow
	 *            whether to apply a drop shadow to the plot area
	 * @see Chart#setPlotShadow(boolean)
	 */
	public void setPlotShadow(boolean plotShadow) {
		_chartsApi.setPlotShadow(plotShadow);
	}

	/**
	 * Returns whether to apply a drop shadow to the plot area. Requires that
	 * {@link #setPlotBackgroundColor(String)} be set.
	 * <p>
	 * Default: false.
	 * 
	 * @see Chart#isPlotShadow()
	 */
	public boolean isPlotShadow() {
		return _chartsApi.isPlotShadow();
	}

	/**
	 * Sets to true to enable the cartesian charts like line, spline, area and
	 * column are transformed into the polar coordinate system.
	 * 
	 * @param polar
	 *            when true, cartesian charts like line, spline, area and column
	 *            are transformed into the polar coordinate system
	 * @see Chart#setPolar(boolean)
	 */
	public void setPolar(boolean polar) {
		_chartsApi.setPolar(polar);
	}

	/**
	 * Returns whether transforms into the polar coordinate system
	 * <p>
	 * Default: false.
	 * 
	 * @see Chart#isPolar()
	 */
	public boolean isPolar() {
		return _chartsApi.isPolar();
	}

	/**
	 * Sets whether to reflow the chart to fit the width of the container div on
	 * resizing the window.
	 * 
	 * @param reflow
	 *            whether to reflow the chart to fit the width of the container
	 *            div on resizing the window
	 * @see Chart#setReflow(boolean)
	 */
	public void setReflow(boolean reflow) {
		_chartsApi.setReflow(reflow);
	}

	/**
	 * Returns whether to reflow the chart to fit the width of the container div
	 * on resizing the window.
	 * <p>
	 * Default: true.
	 * 
	 * @see Chart#isReflow()
	 */
	public boolean isReflow() {
		return _chartsApi.isReflow();
	}

	/**
	 * Sets the background color of the marker square when selecting (zooming in
	 * on) an area of the chart.
	 */
	public void setSelectionMarkerFill(String selectionMarkerFill) {
		_chartsApi.setSelectionMarkerFill(selectionMarkerFill);
	}

	/**
	 * Sets the background color of the marker square when selecting (zooming in
	 * on) an area of the chart.
	 */
	public void setSelectionMarkerFill(LinearGradient selectionMarkerFill) {
		_chartsApi.setSelectionMarkerFill(selectionMarkerFill);
	}

	/**
	 * Sets the background color of the marker square when selecting (zooming in
	 * on) an area of the chart.
	 */
	public void setSelectionMarkerFill(RadialGradient selectionMarkerFill) {
		_chartsApi.setSelectionMarkerFill(selectionMarkerFill);
	}

	/**
	 * Sets the background color of the marker square when selecting (zooming in
	 * on) an area of the chart.
	 */
	public void setSelectionMarkerFill(Color selectionMarkerFill) {
		_chartsApi.setSelectionMarkerFill(selectionMarkerFill);
	}

	/**
	 * Returns the background color of the marker square when selecting (zooming
	 * in on) an area of the chart.
	 * <p>
	 * Default: "rgba(69,114,167,0.25)".
	 * 
	 * @see Chart#getSelectionMarkerFill()
	 */
	public Color getSelectionMarkerFill() {
		return _chartsApi.getSelectionMarkerFill();
	}

	/**
	 * Sets whether to apply a drop shadow to the outer chart area. Requires
	 * that backgroundColor be set.
	 * 
	 * @param shadow
	 *            whether to apply a drop shadow to the outer chart area
	 * @see Chart#setShadow(boolean)
	 */
	public void setShadow(boolean shadow) {
		_chartsApi.setShadow(shadow);
	}

	/**
	 * Returns whether to apply a drop shadow to the outer chart area. Requires
	 * that backgroundColor be set.
	 * <p>
	 * Default: false.
	 * 
	 * @see Chart#isShadow()
	 */
	public boolean isShadow() {
		return _chartsApi.isShadow();
	}

	/**
	 * Sets whether to show the axes initially. This only applies to empty
	 * charts where series are added dynamically, as axes are automatically
	 * added to cartesian series.
	 * 
	 * @param showAxes
	 *            whether to show the axes initially
	 * @see Chart#setShowAxes(boolean)
	 */
	public void setShowAxes(boolean showAxes) {
		_chartsApi.setShowAxes(showAxes);
	}

	/**
	 * Returns whether to show the axes initially. This only applies to empty
	 * charts where series are added dynamically, as axes are automatically
	 * added to cartesian series.
	 * <p>
	 * Default: false.
	 * 
	 * @see Chart#isShowAxes()
	 */
	public boolean isShowAxes() {
		return _chartsApi.isShowAxes();
	}

	/**
	 * Sets the distance between the outer edge of the chart and the content,
	 * like title, legend, axis title or labels. The numbers in the array
	 * designate top, right, bottom and left respectively. Use the methods of
	 * {@link #setSpacingTop(Number)}, {@link #setSpacingRight(Number)},
	 * {@link #setSpacingBottom(Number)} and {@link #setSpacingLeft(Number)} for
	 * shorthand setting of one option.
	 * 
	 * @param spacings
	 *            the distance between the outer edge of the chart and the
	 *            content, like title, legend, axis title or labels
	 * @see #setSpacing(Number[])
	 */
	public void setSpacing(String spacings) {
		_chartsApi.setSpacing(spacings);
	}

	/**
	 * Sets the distance between the outer edge of the chart and the content,
	 * like title, legend, axis title or labels. The numbers in the array
	 * designate top, right, bottom and left respectively. Use the methods of
	 * {@link #setSpacingTop(Number)}, {@link #setSpacingRight(Number)},
	 * {@link #setSpacingBottom(Number)} and {@link #setSpacingLeft(Number)} for
	 * shorthand setting of one option.
	 * 
	 * @param spacing
	 *            the distance between the outer edge of the chart and the
	 *            content, like title, legend, axis title or labels
	 * @see Chart#setSpacing(Number[])
	 */
	public void setSpacing(Number[] spacing) {
		_chartsApi.setSpacing(spacing);
	}

	/**
	 * Returns the distance between the outer edge of the chart and the content,
	 * like title, legend, axis title or labels. The numbers in the array
	 * designate top, right, bottom and left respectively.
	 * <p>
	 * Default: [10, 10, 15, 10].
	 */
	public Number[] getSpacing() {
		return _chartsApi.getSpacing();
	}

	/**
	 * Sets the space between the bottom edge of the chart and the content (plot
	 * area, axis title and labels, title, subtitle or legend in top position).
	 * </p>
	 * 
	 * @param spacingBottom
	 *            The space between the bottom edge of the chart and the content
	 *            (plot area, axis title and labels, title, subtitle or legend
	 *            in top position)
	 * @see Chart#setSpacingBottom(Number)
	 */
	public void setSpacingBottom(Number spacingBottom) {
		_chartsApi.setSpacingBottom(spacingBottom);
	}

	/**
	 * Returns the space between the bottom edge of the chart and the content
	 * (plot area, axis title and labels, title, subtitle or legend in top
	 * position).
	 * <p>
	 * Default: 15.
	 * 
	 * @see Chart#getSpacingBottom()
	 */
	public Number getSpacingBottom() {
		return _chartsApi.getSpacingBottom();
	}

	/**
	 * Sets the space between the left edge of the chart and the content (plot
	 * area, axis title and labels, title, subtitle or legend in top position).
	 * </p>
	 * 
	 * @param spacingLeft
	 *            The space between the left edge of the chart and the content
	 *            (plot area, axis title and labels, title, subtitle or legend
	 *            in top position)
	 * @see Chart#setSpacingLeft(Number)
	 */
	public void setSpacingLeft(Number spacingLeft) {
		_chartsApi.setSpacingLeft(spacingLeft);
	}

	/**
	 * Returns the space between the left edge of the chart and the content
	 * (plot area, axis title and labels, title, subtitle or legend in top
	 * position).
	 * <p>
	 * Default: 10.
	 * 
	 * @see Chart#getSpacingLeft()
	 */
	public Number getSpacingLeft() {
		return _chartsApi.getSpacingLeft();
	}

	/**
	 * Sets the space between the right edge of the chart and the content (plot
	 * area, axis title and labels, title, subtitle or legend in top position).
	 * 
	 * @param spacingRight
	 *            The space between the right edge of the chart and the content
	 *            (plot area, axis title and labels, title, subtitle or legend
	 *            in top position)
	 * @see Chart#setSpacingRight(Number)
	 */
	public void setSpacingRight(Number spacingRight) {
		_chartsApi.setSpacingRight(spacingRight);
	}

	/**
	 * Returns the space between the right edge of the chart and the content
	 * (plot area, axis title and labels, title, subtitle or legend in top
	 * position).
	 * <p>
	 * Default: 10.
	 * 
	 * @see Chart#getSpacingRight()
	 */
	public Number getSpacingRight() {
		return _chartsApi.getSpacingRight();
	}

	/**
	 * Sets the space between the top edge of the chart and the content (plot
	 * area, axis title and labels, title, subtitle or legend in top position).
	 * 
	 * @param spacingTop
	 *            The space between the top edge of the chart and the content
	 *            (plot area, axis title and labels, title, subtitle or legend
	 *            in top position)
	 * @see Chart#setSpacingTop(Number)
	 */
	public void setSpacingTop(Number spacingTop) {
		_chartsApi.setSpacingTop(spacingTop);
	}

	/**
	 * Returns the space between the top edge of the chart and the content (plot
	 * area, axis title and labels, title, subtitle or legend in top position).
	 * <p>
	 * Default: 10.
	 * 
	 * @see Chart#getSpacingTop()
	 */
	public Number getSpacingTop() {
		return _chartsApi.getSpacingTop();
	}

	/**
	 * Sets additional CSS styles to apply inline to the container
	 * <code>div</code>. Note that since the default font styles are applied in
	 * the renderer, it is ignorant of the individual chart options and must be
	 * set globally.
	 * <p>
	 * This method allows <b>simple</b> settings like
	 * <pre>{@code
	 * label.setStyle("fontWeight: 'bold'; color: 'white'; textShadow: '0px 1px 2px black'");
	 * }</pre>
	 * <p>
	 * In the format of CSS key:value, and ";" separating multiple key:value pairs,
	 * if you have more a complex CSS styling, we recommend using {@link #setStyle(Map)}
	 * @param style one or many CSS key:value pairs in a single line string format
	 */
	public void setStyle(String style) {
		_chartsApi.setStyle(style);
	}

	/**
	 * Sets additional CSS styles to apply inline to the container
	 * <code>div</code>. Note that since the default font styles are applied in
	 * the renderer, it is ignorant of the individual chart options and must be
	 * set globally.
	 * <p>
	 * This method allows <b>complex</b> styling like:
	 * <pre>
	 * {@code
	 * Map<String, Object> styles = new HashMap<String, Object>();
	 * styles.put("fill", "#E8E8EA");
	 * styles.put("stroke", "#6E6E78");
	 * styles.put("stroke-width", "1");
	 * styles.put("r", "0");
	 * Map<String, Object> states = new HashMap<String, Object>();
	 * Map<String, String> hover = new HashMap<String, String>();
	 * hover.put("fill", "#F4F4F6");
	 * hover.put("stroke", "#6E6E78");
	 * hover.put("stroke-width", "1");
	 * hover.put("r", "0");
	 * states.put("hover", hover);
	 * styles.put("states", states);
	 * Map<String, String> style = new HashMap<String, String>();
	 * style.put("color", "#4D5663");
	 * style.put("fontSize", "12px");
	 * style.put("fontFamily", "'Arial'");
	 * style.put("fontWeight", "normal");
	 * style.put("padding", "10px");
	 * styles.put("style", style);
	 * }</pre>
	 * which will generate the final results like:
	 * <pre>{@code
	 * {
	 *     "fill": "#E8E8EA",
	 *     "stroke": "#6E6E78",
	 *     "stroke-width": "1",
	 *     "r": "0",
	 *     "states":
	 *     {
	 *         "hover":
	 *         {
	 *             "fill": "#F4F4F6",
	 *             "stroke": "#6E6E78",
	 *             "stroke-width": "1",
	 *             "r": "0"
	 *         }
	 *     },
	 *     "style":
	 *     {
	 *         "color": "#4D5663",
	 *         "fontSize": "12px",
	 *         "fontFamily": "'Arial'",
	 *         "fontWeight": "normal",
	 *         "padding": "10px"
	 *     }
	 * }
	 * }</pre>
	 * @param <K> Type for the key, usually String
	 * @param <V> Type for the value, usually String. For nested maps, use Object or any other type.
	 * @param style the style object
	 */
	public <K, V> void setStyle(Map<K, V> style) {
		_chartsApi.setStyle(style);
	}

	/**
	 * Returns additional CSS styles to apply inline to the container
	 * <code>div</code>. Note that since the default font styles are applied in
	 * the renderer, it is ignorant of the individual chart options and must be
	 * set globally. Defaults to:
	 * 
	 * <pre>
	 * fontFamily: "Lucida Grande", "Lucida Sans Unicode", Verdana, Arial, Helvetica, sans-serif; fontSize: 12px
	 * </pre>
	 * <p>
	 * @param <K> type for the key
	 * <p>
	 * if style was set using {@link #setStyle(String)}, type will be String
	 * <p>
	 * if style was set using {@link #setStyle(Map)}, type will be the same as the type from the key of the map
	 * <p>
	 * if style was not set, default style will be applied, and type will be String
	 * @param <V> type for the value
	 * <p>
	 * if style was set using {@link #setStyle(String)}, type will be String
	 * <p>
	 * if style was set using {@link #setStyle(Map)}, type will be the same as the type from the key of the map
	 * <p>
	 * if style was not set, default style will be applied, and type will be String
	 */
	public <K, V> Map<K, V> getStyle() {
		return _chartsApi.getStyle();
	}

	/**
	 * Sets decides in what dimentions the user can zoom by dragging the mouse.
	 * Can be one of <code>x</code>, <code>y</code> or <code>xy</code>.
	 * 
	 * @param zoomType
	 *            decides in what dimentions the user can zoom by dragging the
	 *            mouse
	 * @see Chart#setZoomType(String)
	 */
	public void setZoomType(String zoomType) {
		_chartsApi.setZoomType(zoomType);
	}

	/**
	 * Returns decides in what dimentions the user can zoom by dragging the
	 * mouse. Can be one of <code>x</code>, <code>y</code> or <code>xy</code>.
	 * <p>
	 * Default: null.
	 * 
	 * @see Chart#getZoomType()
	 */
	public String getZoomType() {
		return _chartsApi.getZoomType();
	}

	/**
	 * mark a draw flag to inform that this Chart needs update.
	 */
	protected void smartDrawChart() {
		if (_smartDrawChart || _model == null) { // already mark smart draw
			return;
		}
		_smartDrawChart = true;
		
		// do it directly
		try {
			doSmartDraw();
		} finally {
			_smartDrawChart = false;
		}
	}

	/**
	 * Returns the Tooltip options, if any. Otherwise, create a new one.
	 * 
	 * @see Tooltip
	 */
	public Tooltip getTooltip() {
		return _chartsApi.getTooltip();
	}

	/**
	 * Sets the Tooltip options
	 * 
	 * @param tooltip
	 */
	public void setTooltip(Tooltip tooltip) {
		_chartsApi.setTooltip(tooltip);
	}

	/**
	 * Returns the Chart options, if any. Otherwise, create a new one.
	 * 
	 * @see Chart
	 */
	public Chart getChart() {
		return _chartsApi.getChart();
	}

	/**
	 * Sets the no-data options
	 * @since 1.1.0
	 */
	public void setNoData(NoData noData) {
		_chartsApi.setNoData(noData);
	}

	/**
	 * Returns the no-data options
	 * @since 1.1.0
	 */
	public NoData getNoData() {
		return _chartsApi.getNoData();
	}
	/**
	 * Returns the Exporting options, if any. Otherwise, create a new one.
	 * 
	 * @see Exporting
	 */
	public Exporting getExporting() {
		return _chartsApi.getExporting();
	}

	/**
	 * Sets the Exporting options.
	 * 
	 * @see Exporting
	 */
	public void setExporting(Exporting exporting) {
		_chartsApi.setExporting(exporting);
	}

	/**
	 * Sets the Credits options
	 * 
	 * @param credits
	 */
	public void setCredits(Credits credits) {
		_chartsApi.setCredits(credits);
	}

	/**
	 * Returns the Credits options, if any. Otherwise, create a new one.
	 * 
	 * @see Credits
	 */
	public Credits getCredits() {
		return _chartsApi.getCredits();
	}

	/**
	 * Sets the Chart options
	 * 
	 * @param chart
	 */
	public void setChart(Chart chart) {
		_chartsApi.setChart(chart);
	}

	/**
	 * Returns the pane at the index 0
	 */
	public Pane getPane() {
		return _chartsApi.getPane();
	}

	/**
	 * Returns the pane from the given index
	 */
	public Pane getPane(int index) {
		return _chartsApi.getPane(index);
	}

	/**
	 * Returns the size of the pane list
	 */
	public int getPaneSize() {
		return _chartsApi.getPaneSize();
	}

	/**
	 * Returns the Legend options, if any. Otherwise, create a new one.
	 * 
	 * @see Legend
	 */
	public Legend getLegend() {
		return _chartsApi.getLegend();
	}

	/**
	 * Returns the Loading options, if any. Otherwise, create a new one.
	 * 
	 * @see Loading
	 */
	public Loading getLoading() {
		return _chartsApi.getLoading();
	}

	/**
	 * Sets the Legend options
	 * 
	 * @param legend
	 */
	public void setLegend(Legend legend) {
		_chartsApi.setLegend(legend);
	}

	/**
	 * Returns the labels options
	 */
	public Labels getLabels() {
		return _chartsApi.getLabels();
	}

	/**
	 * Sets the labels options
	 */
	public void setLabels(Labels labels) {
		_chartsApi.setLabels(labels);
	}

	/**
	 * Sets the navigation options
	 */
	public void setNavigation(Navigation navigation) {
		_chartsApi.setNavigation(navigation);
	}

	/**
	 * Returns the navigation options
	 */
	public Navigation getNavigation() {
		return _chartsApi.getNavigation();
	}

	/**
	 * Returns the plotOptions option, if any. Otherwise, create a new one.
	 * 
	 * @see PlotOptions
	 */
	public PlotOptions getPlotOptions() {
		return _chartsApi.getPlotOptions();
	}

	/**
	 * Sets the plotOptions options
	 * 
	 * @param plotOptions
	 */
	public void setPlotOptions(PlotOptions plotOptions) {
		_chartsApi.setPlotOptions(plotOptions);
	}

	/**
	 * Returns the Series options, if any. Otherwise, create a new one.
	 * <p>
	 * Returns the first series (index 0) by default.
	 * 
	 * @see Series
	 * @see #getSeries(int)
	 */
	public Series getSeries() {
		return _chartsApi.getSeries();
	}

	/**
	 * Add a series to the chart after render time. Note that this method should
	 * never be used when adding data synchronously at chart render time, as it
	 * adds expense to the calculations and rendering. When adding data at the
	 * same time as the chart is initiated, add the series as a configuration
	 * option instead.
	 * 
	 * @param series
	 */
	public void addSeries(Series series) {
		_chartsApi.addSeries(series);
	}

	/**
	 * Add an axis to the chart after render time. Note that this method should
	 * never be used when adding data synchronously at chart render time, as it
	 * adds expense to the calculations and rendering. When adding data at the
	 * same time as the chart is initiated, add the axis as a configuration
	 * option instead.
	 * 
	 * @param axis
	 */
	public void addAxis(Axis axis) {
		_chartsApi.addAxis(axis);
	}

	/**
	 * Returns the size of series
	 */
	public int getSeriesSize() {
		return _chartsApi.getSeriesSize();
	}

	/**
	 * Returns the Series options from the given index, if any. Otherwise,
	 * create a new one.
	 * 
	 * @see Series
	 */
	public Series getSeries(int index) {
		return _chartsApi.getSeries(index);
	}

	/**
	 * Returns the XAxis options, if any. Otherwise, create a new one.
	 * <p>
	 * Returns the first xAxis (index 0) by default.
	 * 
	 * @see XAxis
	 * @see #getXAxis(int)
	 */
	public XAxis getXAxis() {
		return _chartsApi.getXAxis();
	}

	/**
	 * Returns the XAxis options from the given index, if any. Otherwise, create
	 * a new one.
	 * 
	 * @see XAxis
	 */
	public XAxis getXAxis(int index) {
		return _chartsApi.getXAxis(index);
	}

	/**
	 * Returns the size of xAxis
	 */
	public int getXAxisSize() {
		return _chartsApi.getXAxisSize();
	}
	
	/**
	 * Sets the xAxis with out putting it in an Array
	 */
	public void setXAxis(XAxis xaxis) {
		_chartsApi.setXAxis(xaxis);
	}

	/**
	 * Returns the YAxis options, if any. Otherwise, create a new one.
	 * <p>
	 * Returns the first yAxis (index 0) by default.
	 * 
	 * @see YAxis
	 * @see #getYAxis(int)
	 */
	public YAxis getYAxis() {
		return _chartsApi.getYAxis();
	}

	/**
	 * Returns the YAxis options from the given index, if any. Otherwise, create
	 * a new one.
	 * 
	 * @see YAxis
	 */
	public YAxis getYAxis(int index) {
		return _chartsApi.getYAxis(index);
	}

	/**
	 * Returns the size of yAxis
	 */
	public int getyAxisSize() {
		return _chartsApi.getyAxisSize();
	}
	
	/**
	 * Sets the yAxis with out putting it in an Array
	 */
	public void setYAxis(YAxis yaxis) {
		_chartsApi.setYAxis(yaxis);
	}
	
	/**
	 * Returns the ZAxis options, if any. Otherwise, create a new one.
	 * <p>
	 * Returns the first zAxis (index 0) by default.
	 * 
	 * @see ZAxis
	 * @see #getZAxis(int)
	 */
	public ZAxis getZAxis() {
		return _chartsApi.getZAxis();
	}

	/**
	 * Returns the ZAxis options from the given index, if any. Otherwise, create
	 * a new one.
	 * 
	 * @see ZAxis
	 */
	public ZAxis getZAxis(int index) {
		return _chartsApi.getZAxis(index);
	}

	/**
	 * Returns the size of zAxis
	 */
	public int getzAxisSize() {
		return _chartsApi.getzAxisSize();
	}
	
	/**
	 * Sets the zAxis with out putting it in an Array
	 */
	public void setZAxis(ZAxis zaxis) {
		_chartsApi.setZAxis(zaxis);
	}
	
	/**
	 * Returns the options to render charts in 3 dimensions.
	 * <p>
	 * Default: null.
	 */
	public Options3D getOptions3D() {
		return _chartsApi.getOptions3D();
	}

	/**
	 * Sets the options to render charts in 3 dimensions.
	 * 
	 * @param options3D
	 *            the options to render charts in 3 dimensions.
	 */
	public void setOptions3D(Options3D options3D) {
		_chartsApi.setOptions3D(options3D);
	}
	
	/**
	 * Returns the ColorAxis options, if any. Otherwise, create a new one.
	 * <p>
	 * Returns the first colorAxis (index 0) by default.
	 * 
	 * @see ColorAxis
	 * @see #getColorAxis()
	 */
	public ColorAxis getColorAxis() {
		return _chartsApi.getColorAxis();
	}

	public ColorAxis getColorAxis(int index) {
		return _chartsApi.getColorAxis(index);
	}

	/**
	 * Sets the chart title.
	 */
	public void setTitle(String title) {
		_chartsApi.setTitle(title);
	}

	/**
	 * Sets the Title options
	 * 
	 * @param title
	 */
	public void setTitle(Title title) {
		_chartsApi.setTitle(title);
	}

	/**
	 * Returns the Title options, if any. Otherwise, create a new one.
	 * 
	 * @see Title
	 */
	public Title getTitle() {
		return _chartsApi.getTitle();
	}

	/**
	 * Returns the Subtitle options, if any. Otherwise, create a new one.
	 * 
	 * @see Subtitle
	 */
	public Subtitle getSubtitle() {
		return _chartsApi.getSubtitle();
	}

	/**
	 * Sets the chart subtitle
	 * 
	 * @param subtitle
	 */
	public void setSubtitle(String subtitle) {
		_chartsApi.setSubtitle(subtitle);
	}

	/**
	 * Sets the Subtitle options
	 * 
	 * @param subtitle
	 */
	public void setSubtitle(Subtitle subtitle) {
		_chartsApi.setSubtitle(subtitle);
	}

	/**
	 * Sets whether selection event will zoom the chart to the selected area.
	 * 
	 * @see Chart#setEnableZoomSelection(boolean)
	 */
	public void setEnableZoomSelection(boolean enableZoomSelection) {
		_chartsApi.setEnableZoomSelection(enableZoomSelection);
	}

	/**
	 * Returns whether selection event will zoom the chart to the selected area.
	 * <p>
	 * Default: true.
	 * 
	 * @see Chart#isEnableZoomSelection()
	 */
	public boolean isEnableZoomSelection() {
		return _chartsApi.isEnableZoomSelection();
	}

	/**
	 * Returns what keystrokes to intercept.
	 * <p>Default: null.
	 * @since 7.2.1.2
	 */
	public String getCtrlKeys() {
		return _ctrlKeys;
	}

	/**
	 * Sets what keystrokes to intercept.
	 *
	 * <p>The string could be a combination of the following:
	 * <dl>
	 * <dt>^k</dt>
	 * <dd>A control key, i.e., Ctrl+k, where k could be a~z, 0~9, #n</dd>
	 * <dt>@k</dt>
	 * <dd>A alt key, i.e., Alt+k, where k could be a~z, 0~9, #n</dd>
	 * <dt>$n</dt>
	 * <dd>A shift key, i.e., Shift+n, where n could be #n.
	 * Note: $a ~ $z are not supported.</dd>
	 * <dt>#home</dt>
	 * <dd>Home</dd>
	 * <dt>#end</dt>
	 * <dd>End</dd>
	 * <dt>#ins</dt>
	 * <dd>Insert</dd>
	 * <dt>#del</dt>
	 * <dd>Delete</dd>
	 * <dt>#bak</dt>
	 * <dd>Backspace</dd>
	 * <dt>#left</dt>
	 * <dd>Left arrow</dd>
	 * <dt>#right</dt>
	 * <dd>Right arrow</dd>
	 * <dt>#up</dt>
	 * <dd>Up arrow</dd>
	 * <dt>#down</dt>
	 * <dd>Down arrow</dd>
	 * <dt>#pgup</dt>
	 * <dd>PageUp</dd>
	 * <dt>#pgdn</dt>
	 * <dd>PageDn</dd>
	 * <dt>#f1 #f2 ... #f12</dt>
	 * <dd>Function keys representing F1, F2, ... F12</dd>
	 * </dl>
	 *
	 * <p>For example,
	 * <dl>
	 * <dt>^a^d@c#f10#left#right</dt>
	 * <dd>It means you want to intercept Ctrl+A, Ctrl+D, Alt+C, F10,
	 * Left and Right.</dd>
	 * <dt>^#left</dt>
	 * <dd>It means Ctrl+Left.</dd>
	 * <dt>^#f1</dt>
	 * <dd>It means Ctrl+F1.</dd>
	 * <dt>@#f3</dt>
	 * <dd>It means Alt+F3.</dd>
	 * </dl>
	 *
	 * <p>Note: it doesn't support Ctrl+Alt, Shift+Ctrl, Shift+Alt or Shift+Ctrl+Alt.
	 * @since 7.2.1.2
	 */
	public void setCtrlKeys(String ctrlKeys) {
		if (Strings.isEmpty(ctrlKeys))
			ctrlKeys = null;
		if (!Objects.equals(_ctrlKeys, ctrlKeys)) {
			_ctrlKeys = ctrlKeys;
			smartUpdate("ctrlKeys", ctrlKeys);
		}
	}

	private static String WRAP_FUNC = "function () {var self = this;this.%1$s;}";
	private static String RETURN_FUNC = "(function () {return self.%1$s;})()";

	/**
	 * Add a series to the chart as drilldown from a specific point in the
	 * parent series. This method is used for async drilldown, when clicking a
	 * point in a series should result in loading and displaying a more
	 * high-resolution series. When not async, the setup is simpler using the
	 * drilldown.series options structure.
	 * 
	 * @param point
	 *            The existing Point object from which the drilldown will start.
	 * @param series
	 *            The series options, as documented under plotOptions.series and
	 *            under the plotOptions for each series type
	 */
	public void addSeriesAsDrilldown(Point point, Series series) {
		_drilldownLevels.addFirst(series);
		JSFunction returnFunc = new JSFunction();
		returnFunc
				.callArray("data", point.getSeries().getData().indexOf(point));
		returnFunc.callArray("series", 0); // in drilldown mode, at client only
											// has one series
		JSFunction func = new JSFunction();
		func.callFunction(
				"addSeriesAsDrilldown",
				new JavaScriptValue(String.format(RETURN_FUNC,
						returnFunc.reverse())), series);
		smartUpdate("eval",
				new JavaScriptValue(String.format(WRAP_FUNC, func)), true);
	}

	private void doSmartDraw() {
		if (!checkLock()) return;
		
		if (_model == null)
			throw new UiException("Chart Model cannot be null!");

		PlotData plotData = getPlotEngine().drawPlot(_model);
		if (!getPlotData().equals(plotData))
			_plotData.merge(plotData);
		smartUpdate("plotData", _plotData);
	}

	private transient PlotEngine _chartImpl; // chart implementation

	private PlotEngine getPlotEngine() {
		if (_chartImpl != null)
			return _chartImpl;

		_chartImpl = PlotEngineImpl.getPlotEngine(getType(), getPlotData());
		return _chartImpl;
	}

	private boolean _rendered;

	/**
	 * Returns whether the component has been rendered to client side. Most of
	 * the time this method is used for component developer only.
	 */
	public boolean isRendered() {
		return _rendered;
	}

	// override
	public void invalidate() {
		super.invalidate();
		_rendered = false; // reset
		if (_drilldownLevels != null)
			_drilldownLevels.clear();
	}

	// override
	protected void renderProperties(org.zkoss.zk.ui.sys.ContentRenderer renderer)
			throws java.io.IOException {
		_rendered = true;
		super.renderProperties(renderer);

		// we shoulld use _befInvoke at the first
		if (!_befInvoke.isEmpty()) {
			render(renderer, "eval", _befInvoke);
			_befInvoke.clear();
		}

		if (_plotData != null)
			render(renderer, "plotData", _plotData);

		if (_options != null)
			render(renderer, "options", _options); // just in case

		if (_theme != null)
			render(renderer, "theme", _theme);

		render(renderer, "zclass", _zclass);
		
		render(renderer, "hflex", _hflex);
		render(renderer, "vflex", _vflex);
		render(renderer, "ctrlKeys", _ctrlKeys);
	}

	private LinkedList<Series> _drilldownLevels;

	protected void smartUpdate(String attr, Object value, boolean append) {
		super.smartUpdate(attr, value, append);
	}

	@SuppressWarnings("rawtypes")
	public void service(AuRequest request, boolean everError) {
		final String cmd = request.getCommand();
		final Map data = request.getData();

		if (cmd.equals("dummyEcho")) {
			String attrId = (String) data.get("id");
			EventListener listener = (EventListener) removeAttribute(attrId);
			try {
				listener.onEvent(new Event("onCallBack", this, data
						.get("value")));
			} catch (Exception e) {
				e.printStackTrace();
			}
		} else if (cmd.equals("onDomEvent")) {
			String attrId = (String) data.get("id");
			String name = (String) data.get("name");
			Map dataMap = (Map) data.get("data");

			final int keys = AuRequests.parseKeys(dataMap);
			MouseEvent evt = new MouseEvent(name, this,
					AuRequests.getInt(dataMap, "x", 0, true),
					AuRequests.getInt(dataMap, "y", 0, true),
					AuRequests.getInt(dataMap, "pageX", 0, true),
					AuRequests.getInt(dataMap, "pageY", 0, true), keys);
			Iterable<EventListener<MouseEvent>> list = (Iterable<EventListener<MouseEvent>>) getAttribute(attrId);
			try {
				if (list != null) {
					for (Iterator<EventListener<MouseEvent>> it = list.iterator(); it.hasNext();) {
						it.next().onEvent(evt);
					}
				}
			} catch (Exception e) {
				e.printStackTrace();
			}
		} else if (cmd.startsWith("onPlot")) {
			ChartsEvent event = ChartsEvent.getChartsEvent(request);
			if (ChartsEvents.ON_PLOT_DRILL_UP.equals(cmd)) {
				_drilldownLevels.removeFirst();
				final Series first = _drilldownLevels.getFirst();
				ChartsEvent evt = new ChartsEvent(request.getCommand(),
						request.getComponent(), -1, -1, null, false, false) {
					public Series getSeries() {
						return first;
					}
				};
				Events.postEvent(evt);
				return;
			} else if (ChartsEvents.ON_PLOT_DRILL_DOWN.equals(cmd)) {
				int sIndex = AuRequests.getInt(data, "sIndex", -1);
				ChartsEvent evt = null;
				if (sIndex == 0) {// first time drilldown
					evt = event;
				} else {
					final Series first = _drilldownLevels.getFirst();
					evt = new ChartsEvent(request.getCommand(),
							request.getComponent(), -1, AuRequests.getInt(data,
									"pIndex", -1),
							(Comparable<?>) data.get("category"), false, false) {
						public Series getSeries() {
							return first;
						}
					};
				}

				if (_drilldownLevels == null) {
					_drilldownLevels = new LinkedList<Series>();
					_drilldownLevels.addFirst(evt.getSeries());
				}
				final int size = _drilldownLevels.size();

				Events.postEvent(evt);
				// user may use drilldown options
				// so we need to put the drilldown series in the
				// _drilldownLevels list
				if (_drilldownLevels.size() == size) {
					final String name = evt.getPoint().getDrilldown();

					for (Series seri : getDrilldown().getSeries()) {
						if (name.equals(seri.getId())) {
							_drilldownLevels.addFirst(seri);
							break;
						}
					}
				}
				return;
			} else if (ChartsEvents.ON_PLOT_SELECT.equals(cmd)) {
				syncSelectedPoints(event.getPoint(), true);
			} else if (ChartsEvents.ON_PLOT_UNSELECT.equals(cmd)) {
				syncSelectedPoints(event.getPoint(), false);
			} else if (ChartsEvents.ON_PLOT_HIDE.equals(cmd)) {
				event.getSeries().setVisibleDirectly(false);
			} else if (ChartsEvents.ON_PLOT_SHOW.equals(cmd)) {
				event.getSeries().setVisibleDirectly(true);
			}
			Events.postEvent(event);
		} else if (Events.ON_CLICK.equals(cmd)) {
			Events.postEvent(ChartsClickEvent.getChartsClickEvent(request));
		} else if (Events.ON_SELECTION.equals(cmd)) {
			Events.postEvent(ChartsSelectionEvent
					.getChartsSelectionEvent(request));
		} else if (Events.ON_OK.equals(cmd) || Events.ON_CANCEL.equals(cmd) || Events.ON_CTRL_KEY.equals(cmd)) {
			Events.postEvent(KeyEvent.getKeyEvent(request));
		} else {
			super.service(request, everError);
		}
	}
	
	private void syncSelectedPoints(Point selPoint, boolean selected) {
		disableClientUpdate(true);
		try {
			selPoint.setSelectedDirectly(selected);
			if (selected) {
				_selPoints.add(selPoint);
			} else {
				_selPoints.remove(selPoint);
			}
		} finally {
			disableClientUpdate(false);
		}
	}
	
	public void selectPoint(Point selPoint, boolean accumulate) {
		disableClientUpdate(true);
		try {
			_selPoints.remove(selPoint);
			if (!accumulate) {
				for (Point p: _selPoints) {
					p.setSelectedDirectly(false);
				}
				_selPoints.clear();
			}
			if (selPoint.isSelected())
				_selPoints.add(selPoint);
		} finally {
			disableClientUpdate(false);
		}
	}
	
	/**
	 * Returns a readonly set of all currently selected points in the chart.
	 * Points can be selected by {@link Point#setSelected(boolean)}, {@link Point#select(boolean, boolean)} or by clicking.
	 * <p>
	 * Note: the returned set is not live data. You need to call {@link #getSelectedPoints()} again to 
	 * retrieve the modified set when changing the point's selected attribute. 
	 */
	public Set<Point> getSelectedPoints() {
		return Collections.unmodifiableSet(new HashSet<Point>(_selPoints));
	}

	/**
	 * Clears the selected points in the chart.
	 */
	public void clearPointSelection() {
		if (_selPoints.isEmpty())
			return;
		for (Point p: getSelectedPoints()) {
			p.setSelected(false);
		}
	}

	/**
	 * Sets the zclass for this component
	 * 
	 * @param zclass
	 */
	public void setZclass(String zclass) {
		if (!(Objects.equals(_zclass, zclass))) {
			_zclass = zclass;
			smartUpdate("zclass", zclass);
		}
	}

	/**
	 * The default zclass is "z-charts"
	 */
	public String getZclass() {
		return (this._zclass != null ? this._zclass : "z-charts");
	}
	
	/**
	 * Returns horizontal flex hint of this chart.
	 * <p>Default: null
	 * @return horizontal flex hint of this chart.
	 * @see #setHflex 
	 */
	public String getHflex() {
		return _hflex;
	}
	
	/**
	 * Sets horizontal flex hint of this chart.
	 * <p>Default: null
	 * @see #getHflex 
	 */
	public void setHflex(String flex) {
		if ("min".equals(flex)) {
			throw new WrongValueException("Charts do not support min flex");
		} else if (flex != null && flex.length() == 0) {
			flex = null;
		}
		if (!Objects.equals(_hflex, flex)) {
			_hflex = flex;
			smartUpdate("hflex", flex);
		}
	}
	
	/**
	 * Returns vertical flex hint of this chart.
	 * <p>Default: null
	 * @return vertical flex hint of this chart.
	 * @see #setVflex 
	 */
	public String getVflex() {
		return _vflex;
	}
	
	/**
	 * Sets vertical flex hint of this chart.
	 * <p>Default: null
	 * @see #getVflex 
	 */
	public void setVflex(String flex) {
		if ("min".equals(flex)) {
			throw new WrongValueException("Charts do not support min flex");
		} else if (flex != null && flex.length() == 0) {
			flex = null;
		}
		if (!Objects.equals(_vflex, flex)) {
			_vflex = flex;
			smartUpdate("vflex", flex);
		}
	}

	public Annotation getAnnotation() {
		return _chartsApi.getAnnotation();
	}

	public Annotation getAnnotation(int index) {
		return _chartsApi.getAnnotation(index);
	}

	public int getAnnotationSize() {
		return _chartsApi.getAnnotationSize();
	}

	public void addAnnotation(Annotation annotation) {
		_chartsApi.addAnnotation(annotation);
	}

	public void setCaption(String caption) {
		_chartsApi.setCaption(caption);
	}

	public void setCaption(Caption caption) {
		_chartsApi.setCaption(caption);
	}

	public Caption getCaption() {
		return _chartsApi.getCaption();
	}

	public Responsive getResponsive() {
		return _chartsApi.getResponsive();
	}

	public void setResponsive(Responsive responsive) {
		_chartsApi.setResponsive(responsive);
	}
}
