/* ChartsDetached.java

	Purpose:
		
	Description:
		
	History:
		Tue Dec 19 12:10:13 CST 2017, Created by rudyhuang

Copyright (C) 2017 Potix Corporation. All Rights Reserved.
 */
package org.zkoss.chart;

import static org.zkoss.chart.Charts.POLAR;

import java.io.Serializable;
import java.util.Arrays;
import java.util.List;
import java.util.Map;

import org.zkoss.chart.impl.PlotEngineImpl;
import org.zkoss.chart.model.ChartsModel;
import org.zkoss.chart.model.ChartsModelProxy;
import org.zkoss.chart.options3D.Options3D;
import org.zkoss.chart.plotOptions.PlotOptions;
import org.zkoss.lang.Objects;
import org.zkoss.zk.ui.UiException;
import org.zkoss.zk.ui.WrongValueException;
import org.zkoss.zul.ChartModel;
import org.zkoss.zul.impl.Utils;

/**
 * An engine of {@link Charts}. Does not need a ZK execution to work.
 * It is used mainly to generate options as JSON format for
 * <a href="https://export.highcharts.com/">Highcharts Export Server</a>.
 * Typical usage:
 * <pre>{@code
 * ChartsEngine chart = new ChartsEngine();
 * chart.setTitle("Monthly Average Temperature");
 * chart.setSubTitle("Source: WorldClimate.com");
 * chart.setType(Charts.LINE);
 * chart.setTheme(Theme.DEFAULT);
 * chart.setModel(model);
 * chart.getXAxis().setCategories(Arrays.asList(new String[]{
 *     "Jan", "Feb", "Mar", "Apr", "May", "Jun",
 *     "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
 * }));
 * chart.getYAxis().getTitle().setText("Temperature (°C)");
 * return chart.toJSON(); // {"xAxis":{}...}
 * }</pre>
 *
 * @see #toJSON()
 * @author rudyhuang
 * @since 3.0.2
 */
public class ChartsEngine implements ChartsApi, Serializable {

	private static final long serialVersionUID = 20171219150545L;

	private boolean _smartDrawChart; // whether post the smartDraw event already?
	private ChartsModel _model; // chart data model
	private PlotData _plotData;
	private Theme _theme;

	private final Charts _owner;

	public ChartsEngine() {
		this(null);
	}

	// For Charts class
	ChartsEngine(Charts owner) {
		_owner = owner;
	}

	public PlotData getPlotData() {
		if (_plotData == null) {
			_plotData = new PlotData(_owner);
			if (_owner != null) _owner.getPlotData(); // sync Charts' _plotData
		}
		return _plotData;
	}

	public Drilldown getDrilldown() {
		return getPlotData().getDrilldown();
	}

	public void setColors(List<Color> colors) {
		getPlotData().setColors(colors);
	}

	public void setColors(String... source) {
		Color[] colors = new Color[source.length];
		int i = 0;
		for (String s : source)
			colors[i++] = new Color(s);
		setColors(Arrays.asList(colors));
	}

	public List<Color> getColors() {
		return getPlotData().getColors();
	}

	public void setTheme(Theme theme) {
		if (theme == null)
			throw new WrongValueException("Theme options cannot be null!");

		_theme = theme; // no need to check, force to update
		String[] themeColors = ThemeColors.getThemeColors(theme);
		if (themeColors != null) {
			setColors(themeColors);
		}
	}

	public Theme getTheme() {
		return _theme;
	}

	public ChartsModel getModel() {
		return _model;
	}

	public void setModel(ChartsModel model) {
		if (_model != model) {
			_model = model;
			smartDrawChart();
		}
	}

	public void setModel(ChartModel model) {
		if (model != null) {
			// We cannot know the ZK version if we have no execution, so just try it out
			setModel(ChartsModelProxy.getInstance(model));
		} else {
			setModel((ChartsModel) null);
		}
	}

	public void setType(String type) {
		Chart chart = getChart();
		String oldType = chart.getType();
		if (!Objects.equals(oldType, type)) {
			// Highcharts JS didn't have a type for Polar, so we need to do some
			// tricks.
			if (POLAR.equals(type))
				chart.setPolar(true); // just in case
			if (POLAR.equals(oldType))
				chart.setPolar(false);

			chart.setType(type);
			_chartImpl = null; // reset if changed
			smartDrawChart();
		}
	}

	public String getType() {
		return getChart().getType();
	}

	public void setAlignTicks(boolean alignTicks) {
		getChart().setAlignTicks(alignTicks);
	}

	public boolean isAlignTicks() {
		return getChart().isAlignTicks();
	}

	public void setAnimation(boolean animation) {
		getChart().setAnimation(animation);
	}

	public boolean isAnimation() {
		return getChart().isAnimation();
	}

	public void setIgnoreHiddenSeries(boolean ignoreHiddenSeries) {
		getChart().setIgnoreHiddenSeries(ignoreHiddenSeries);
	}

	public boolean isIgnoreHiddenSeries() {
		return getChart().isIgnoreHiddenSeries();
	}

	public void setBackgroundColor(String backgroundColor) {
		getChart().setBackgroundColor(backgroundColor);
	}

	public void setBackgroundColor(LinearGradient gradient) {
		getChart().setBackgroundColor(gradient);
	}

	public void setBackgroundColor(RadialGradient gradient) {
		getChart().setBackgroundColor(gradient);
	}

	public void setBackgroundColor(Color color) {
		getChart().setBackgroundColor(color);
	}

	public Color getBackgroundColor() {
		return getChart().getBackgroundColor();
	}

	public void setBorderColor(String borderColor) {
		getChart().setBorderColor(borderColor);
	}

	public void setBorderColor(LinearGradient borderColor) {
		getChart().setBorderColor(borderColor);
	}

	public void setBorderColor(RadialGradient borderColor) {
		getChart().setBorderColor(borderColor);
	}

	public void setBorderColor(Color borderColor) {
		getChart().setBorderColor(borderColor);
	}

	public Color getBorderColor() {
		return getChart().getBorderColor();
	}

	public void setBorderRadius(Number borderRadius) {
		getChart().setBorderRadius(borderRadius);
	}

	public Number getBorderRadius() {
		return getChart().getBorderRadius();
	}

	public void setBorderWidth(Number borderWidth) {
		getChart().setBorderWidth(borderWidth);
	}

	public Number getBorderWidth() {
		return getChart().getBorderWidth();
	}

	public void setWidth(Number width) {
		getChart().setWidth(width);
	}

	public Number getWidth() {
		return getChart().getWidth();
	}

	public void setHeight(Number height) {
		getChart().setHeight(height);
	}

	public void setHeight(String height) {
		getChart().setHeight(height);
	}

	public Number getHeight() {
		return getChart().getHeight();
	}

	public void setInverted(boolean inverted) {
		getChart().setInverted(inverted);
	}

	public boolean isInverted() {
		return getChart().isInverted();
	}

	public void setMargin(String margins) {
		final int[] margin = Utils.stringToInts(margins, 0);
		Number[] result = new Number[margin.length];
		int i = 0;
		for (int m : margin) {
			result[i++] = m;
		}
		setMargin(result);
	}
	
	public void setMargin(Number margin) {
		getChart().setMargin(margin);
	}

	public void setMargin(Number[] margin) {
		getChart().setMargin(margin);
	}

	public Number[] getMargin() {
		return getChart().getMargin();
	}

	public void setMarginBottom(Number marginBottom) {
		getChart().setMarginBottom(marginBottom);
	}

	public Number getMarginBottom() {
		return getChart().getMarginBottom();
	}

	public void setMarginLeft(Number marginLeft) {
		getChart().setMarginLeft(marginLeft);
	}

	public Number getMarginLeft() {
		return getChart().getMarginLeft();
	}

	public void setMarginRight(Number marginRight) {
		getChart().setMarginRight(marginRight);
	}

	public Number getMarginRight() {
		return getChart().getMarginRight();
	}

	public void setMarginTop(Number marginTop) {
		getChart().setMarginTop(marginTop);
	}

	public Number getMarginTop() {
		return getChart().getMarginTop();
	}

	public void setPinchType(String pinchType) {
		getChart().setPinchType(pinchType);
	}

	public String getPinchType() {
		return getChart().getPinchType();
	}

	public void setPlotBackgroundColor(String plotBackgroundColor) {
		getChart().setPlotBackgroundColor(plotBackgroundColor);
	}

	public void setPlotBackgroundColor(LinearGradient plotBackgroundColor) {
		getChart().setPlotBackgroundColor(plotBackgroundColor);
	}

	public void setPlotBackgroundColor(RadialGradient plotBackgroundColor) {
		getChart().setPlotBackgroundColor(plotBackgroundColor);
	}

	public void setPlotBackgroundColor(Color plotBackgroundColor) {
		getChart().setPlotBackgroundColor(plotBackgroundColor);
	}

	public Color getPlotBackgroundColor() {
		return getChart().getPlotBackgroundColor();
	}

	public void setPlotBackgroundImage(String plotBackgroundImage) {
		getChart().setPlotBackgroundImage(plotBackgroundImage);
	}

	public String getPlotBackgroundImage() {
		return getChart().getPlotBackgroundImage();
	}

	public void setPlotBorderColor(String plotBorderColor) {
		getChart().setPlotBorderColor(plotBorderColor);
	}

	public void setPlotBorderColor(LinearGradient plotBorderColor) {
		getChart().setPlotBorderColor(plotBorderColor);
	}

	public void setPlotBorderColor(RadialGradient plotBorderColor) {
		getChart().setPlotBorderColor(plotBorderColor);
	}

	public void setPlotBorderColor(Color plotBorderColor) {
		getChart().setPlotBorderColor(plotBorderColor);
	}

	public Color getPlotBorderColor() {
		return getChart().getPlotBorderColor();
	}

	public void setPlotBorderWidth(Number plotBorderWidth) {
		getChart().setPlotBorderWidth(plotBorderWidth);
	}

	public Number getPlotBorderWidth() {
		return getChart().getPlotBorderWidth();
	}

	public void setPlotShadow(boolean plotShadow) {
		getChart().setPlotShadow(plotShadow);
	}

	public boolean isPlotShadow() {
		return getChart().isPlotShadow();
	}

	public void setPolar(boolean polar) {
		getChart().setPolar(polar);
	}

	public boolean isPolar() {
		return getChart().isPolar();
	}

	public void setReflow(boolean reflow) {
		getChart().setReflow(reflow);
	}

	public boolean isReflow() {
		return getChart().isReflow();
	}

	public void setSelectionMarkerFill(String selectionMarkerFill) {
		getChart().setSelectionMarkerFill(selectionMarkerFill);
	}

	public void setSelectionMarkerFill(LinearGradient selectionMarkerFill) {
		getChart().setSelectionMarkerFill(selectionMarkerFill);
	}

	public void setSelectionMarkerFill(RadialGradient selectionMarkerFill) {
		getChart().setSelectionMarkerFill(selectionMarkerFill);
	}

	public void setSelectionMarkerFill(Color selectionMarkerFill) {
		getChart().setSelectionMarkerFill(selectionMarkerFill);
	}

	public Color getSelectionMarkerFill() {
		return getChart().getSelectionMarkerFill();
	}

	public void setShadow(boolean shadow) {
		getChart().setShadow(shadow);
	}

	public boolean isShadow() {
		return getChart().isShadow();
	}

	public void setShowAxes(boolean showAxes) {
		getChart().setShowAxes(showAxes);
	}

	public boolean isShowAxes() {
		return getChart().isShowAxes();
	}

	public void setSpacing(String spacings) {
		final int[] spacing = Utils.stringToInts(spacings, 0);
		int i = 0;
		Number[] result = new Number[spacing.length];
		for (int s : spacing) {
			result[i++] = s;
		}
		setSpacing(result);
	}

	public void setSpacing(Number[] spacing) {
		getChart().setSpacing(spacing);
	}

	public Number[] getSpacing() {
		return getChart().getSpacing();
	}

	public void setSpacingBottom(Number spacingBottom) {
		getChart().setSpacingBottom(spacingBottom);
	}

	public Number getSpacingBottom() {
		return getChart().getSpacingBottom();
	}

	public void setSpacingLeft(Number spacingLeft) {
		getChart().setSpacingLeft(spacingLeft);
	}

	public Number getSpacingLeft() {
		return getChart().getSpacingLeft();
	}

	public void setSpacingRight(Number spacingRight) {
		getChart().setSpacingRight(spacingRight);
	}

	public Number getSpacingRight() {
		return getChart().getSpacingRight();
	}

	public void setSpacingTop(Number spacingTop) {
		getChart().setSpacingTop(spacingTop);
	}

	public Number getSpacingTop() {
		return getChart().getSpacingTop();
	}

	public void setStyle(String style) {
		getChart().setStyle(style);
	}

	public <K, V> void setStyle(Map<K, V> style) {
		getChart().setStyle(style);
	}

	public <K, V> Map<K, V> getStyle() {
		return getChart().getStyle();
	}

	public void setZoomType(String zoomType) {
		getChart().setZoomType(zoomType);
	}

	public String getZoomType() {
		return getChart().getZoomType();
	}

	/**
	 * mark a draw flag to inform that this Chart needs update.
	 */
	protected void smartDrawChart() {
		if (_smartDrawChart || _model == null) { // already mark smart draw
			return;
		}
		_smartDrawChart = true;
		
		// do it directly
		try {
			doSmartDraw();
		} finally {
			_smartDrawChart = false;
		}
	}

	public Tooltip getTooltip() {
		return getPlotData().getTooltip();
	}

	public void setTooltip(Tooltip tooltip) {
		getPlotData().setTooltip(tooltip);
	}

	public Chart getChart() {
		return getPlotData().getChart();
	}

	public void setNoData(NoData noData) {
		getPlotData().setNoData(noData);
	}

	public NoData getNoData() {
		return getPlotData().getNoData();
	}
	public Exporting getExporting() {
		return getPlotData().getExporting();
	}

	public void setExporting(Exporting exporting) {
		getPlotData().setExporting(exporting);
	}

	public void setCredits(Credits credits) {
		getPlotData().setCredits(credits);
	}

	public Credits getCredits() {
		return getPlotData().getCredits();
	}

	public void setChart(Chart chart) {
		getPlotData().setChart(chart);
	}

	public Pane getPane() {
		return getPlotData().getPane();
	}

	public Pane getPane(int index) {
		return getPlotData().getPane(index);
	}

	public int getPaneSize() {
		return getPlotData().getPaneSize();
	}

	public Legend getLegend() {
		return getPlotData().getLegend();
	}

	public Loading getLoading() {
		return getPlotData().getLoading();
	}

	public void setLegend(Legend legend) {
		getPlotData().setLegend(legend);
	}

	public Labels getLabels() {
		return getPlotData().getLabels();
	}

	public void setLabels(Labels labels) {
		getPlotData().setLabels(labels);
	}

	public void setNavigation(Navigation navigation) {
		getPlotData().setNavigation(navigation);
	}

	public Navigation getNavigation() {
		return getPlotData().getNavigation();
	}

	public PlotOptions getPlotOptions() {
		return getPlotData().getPlotOptions();
	}

	public void setPlotOptions(PlotOptions plotOptions) {
		getPlotData().setPlotOptions(plotOptions);
	}

	public Series getSeries() {
		return getPlotData().getSeries();
	}

	public void addSeries(Series series) {
		getPlotData().addSeries(series);
	}

	public void addAxis(Axis axis) {
		if (axis instanceof XAxis)
			getPlotData().addXAxis((XAxis) axis);
		else if (axis instanceof YAxis)
			getPlotData().addYAxis((YAxis) axis);
	}

	public int getSeriesSize() {
		return getPlotData().getSeriesSize();
	}

	public Series getSeries(int index) {
		return getPlotData().getSeries(index);
	}

	public XAxis getXAxis() {
		return getPlotData().getXAxis();
	}

	public XAxis getXAxis(int index) {
		return getPlotData().getXAxis(index);
	}

	public int getXAxisSize() {
		return getPlotData().getXAxisSize();
	}
	
	public void setXAxis(XAxis xaxis) {
		getPlotData().setXAxis(xaxis);
	}

	public YAxis getYAxis() {
		return getPlotData().getYAxis();
	}

	public YAxis getYAxis(int index) {
		return getPlotData().getYAxis(index);
	}

	public int getyAxisSize() {
		return getPlotData().getYAxisSize();
	}
	
	public void setYAxis(YAxis yaxis) {
		getPlotData().setYAxis(yaxis);
	}
	
	public ZAxis getZAxis() {
		return getPlotData().getZAxis();
	}

	public ZAxis getZAxis(int index) {
		return getPlotData().getZAxis(index);
	}

	public int getzAxisSize() {
		return getPlotData().getZAxisSize();
	}
	
	public void setZAxis(ZAxis zaxis) {
		getPlotData().setZAxis(zaxis);
	}
	
	public Options3D getOptions3D() {
		return getChart().getOptions3D();
	}

	public void setOptions3D(Options3D options3D) {
		getChart().setOptions3D(options3D);
	}
	
	public ColorAxis getColorAxis() {
		return getPlotData().getColorAxis();
	}

	public ColorAxis getColorAxis(int index) {
		return getPlotData().getColorAxis(index);
	}

	public void setTitle(String title) {
		getTitle().setText(title);
	}

	public void setTitle(Title title) {
		getPlotData().setTitle(title);
	}

	public Title getTitle() {
		return getPlotData().getTitle();
	}

	public Subtitle getSubtitle() {
		return getPlotData().getSubtitle();
	}

	public void setSubtitle(String subtitle) {
		getSubtitle().setText(subtitle);
	}

	public void setSubtitle(Subtitle subtitle) {
		getPlotData().setSubtitle(subtitle);
	}

	public void setEnableZoomSelection(boolean enableZoomSelection) {
		getChart().setEnableZoomSelection(enableZoomSelection);
	}

	public boolean isEnableZoomSelection() {
		return getChart().isEnableZoomSelection();
	}

	private void doSmartDraw() {
		if (_model == null)
			throw new UiException("Chart Model cannot be null!");

		PlotData plotData = getPlotEngine().drawPlot(_model);
		if (!getPlotData().equals(plotData))
			_plotData.merge(plotData);
	}

	private transient PlotEngine _chartImpl; // chart implementation

	private PlotEngine getPlotEngine() {
		if (_chartImpl != null)
			return _chartImpl;

		_chartImpl = PlotEngineImpl.getPlotEngine(getType(), getPlotData());
		return _chartImpl;
	}

	/**
	 * Gets the JSON result.
	 * @return JSON string
	 */
	public String toJSON() {
		return getPlotData().toJSONString();
	}

	public Annotation getAnnotation() {
		return getPlotData().getAnnotations();
	}

	public Annotation getAnnotation(int index) {
		return getPlotData().getAnnotations(index);
	}

	public int getAnnotationSize() {
		return getPlotData().getAnnotationsSize();
	}

	public void addAnnotation(Annotation annotation) {
		getPlotData().addAnnotations(annotation);
	}

	public void setCaption(String caption) {
		getCaption().setText(caption);
	}

	public void setCaption(Caption caption) {
		getPlotData().setCaption(caption);
	}

	public Caption getCaption() {
		return getPlotData().getCaption();
	}

	public Responsive getResponsive() {
		return getPlotData().getResponsive();
	}

	public void setResponsive(Responsive responsive) {
		getPlotData().setResponsive(responsive);
	}
}
