/* Node.java

	Purpose:
		
	Description:
		
	History:
		Fri Nov 15 14:42:02 CST 2019, Created by rudyhuang

Copyright (C) 2019 Potix Corporation. All Rights Reserved.
*/
package org.zkoss.chart;

import org.zkoss.chart.plotOptions.NetworkGraphDataLabels;
import org.zkoss.chart.util.DynamicalAttribute;

/**
 * A collection of options for the individual nodes.
 * The nodes in a diagram are auto-generated instances of Highcharts.Point,
 * but options can be applied here and linked by the id.
 *
 * @author rudyhuang
 * @since 7.2.1.0
 */
public class Node extends Optionable {
	private enum Attrs implements PlotAttribute, DynamicalAttribute {
		borderRadius, color, colorIndex, column, dataLabels, description, id, image, layout,
		level, marker, mass, name, offset, title
	}

	/**
	 * Returns the border radius of the node.
	 */
	public Object getBorderRadius() {
		return getAttr(Attrs.borderRadius, null).asNumber();
	}

	/**
	 * Sets the border radius of the node.
	 */
	public void setBorderRadius(Number borderRadius) {
		setAttr(Attrs.borderRadius, borderRadius);
	}

	/**
	 * Sets the border radius of the node.
	 */
	public void setBorderRadius(String borderRadius) {
		setAttr(Attrs.borderRadius, borderRadius);
	}

	/**
	 * Returns the color of the auto generated node.
	 * <p>
	 * Default: null
	 */
	public String getColor() {
		return getAttr(Attrs.color, null).asString();
	}

	/**
	 * Sets the color of the auto generated node.
	 */
	public void setColor(String color) {
		setAttr(Attrs.color, color);
	}

	/**
	 * Returns the color index of the auto generated node, especially for use in styled mode.
	 * <p>
	 * Default: null
	 */
	public Number getColorIndex() {
		return getAttr(Attrs.colorIndex, null).asNumber();
	}

	/**
	 * Sets the color index of the auto generated node, especially for use in styled mode.
	 */
	public void setColorIndex(Number colorIndex) {
		setAttr(Attrs.colorIndex, colorIndex);
	}

	/**
	 * Returns an optional column index of where to place the node.
	 * The default behaviour is to place it next to the preceding node.
	 * Note that this option name is counter intuitive in inverted charts,
	 * like for example an organization chart rendered top down.
	 * In this case the "columns" are horizontal.
	 * <p>
	 * Default: null
	 */
	public Number getColumn() {
		return getAttr(Attrs.column, null).asNumber();
	}

	/**
	 * Sets an optional column index of where to place the node.
	 * The default behaviour is to place it next to the preceding node.
	 * Note that this option name is counter intuitive in inverted charts,
	 * like for example an organization chart rendered top down.
	 * In this case the "columns" are horizontal.
	 */
	public void setColumn(Number column) {
		setAttr(Attrs.column, column);
	}

	/**
	 * Returns the individual data label for each node, if any. Otherwise, create a new one.
	 * The options are the same as the ones for {@link Series#getDataLabels()}.
	 * <p>
	 * Default: null
	 */
	public NetworkGraphDataLabels getDataLabels() {
		NetworkGraphDataLabels dataLabels = (NetworkGraphDataLabels) this.getAttr(Attrs.dataLabels);
		if (dataLabels == null) {
			dataLabels = new NetworkGraphDataLabels();
			setDataLabels(dataLabels);
		}
		return dataLabels;
	}

	/**
	 * Sets the individual data label for each node.
	 * The options are the same as the ones for {@link Series#getDataLabels()}.
	 */
	public void setDataLabels(NetworkGraphDataLabels dataLabels) {
		setAttr(Attrs.dataLabels, dataLabels);
	}

	/**
	 * Returns the job description for the node card, will be inserted by the default dataLabel.nodeFormatter.
	 * <p>
	 * Default: null
	 */
	public String getDescription() {
		return getAttr(Attrs.description, null).asString();
	}

	/**
	 * Sets the job description for the node card, will be inserted by the default dataLabel.nodeFormatter.
	 */
	public void setDescription(String description) {
		setAttr(Attrs.description, description);
	}

	/**
	 * Returns the id of the auto-generated node, referring to the from or to setting of the link.
	 * <p>
	 * Default: null
	 */
	public String getId() {
		return getAttr(Attrs.id, null).asString();
	}

	/**
	 * Sets the id of the auto-generated node, referring to the from or to setting of the link.
	 */
	public void setId(String id) {
		setAttr(Attrs.id, id);
	}

	/**
	 * Returns an image for the node card, will be inserted by the default dataLabel.nodeFormatter.
	 * <p>
	 * Default: null
	 */
	public String getImage() {
		return getAttr(Attrs.image, null).asString();
	}

	/**
	 * Sets an image for the node card, will be inserted by the default dataLabel.nodeFormatter.
	 */
	public void setImage(String image) {
		setAttr(Attrs.image, image);
	}

	/**
	 * Returns the layout for the node's children.
	 * If {@code hanging}, this node's children will hang below their parent,
	 * allowing a tighter packing of nodes in the diagram.
	 * <p>
	 * Default: normal
	 */
	public String getLayout() {
		return getAttr(Attrs.layout, "normal").asString();
	}

	/**
	 * Sets the layout for the node's children.
	 * If {@code hanging}, this node's children will hang below their parent,
	 * allowing a tighter packing of nodes in the diagram.
	 */
	public void setLayout(String layout) {
		if (!("normal".equals(layout) || "hanging".equals(layout)))
			throw new IllegalArgumentException("Unsupported layout: [" + layout + "]");
		setAttr(Attrs.layout, layout);
	}

	/**
	 * Returns an optional level index of where to place the node.
	 * The default behaviour is to place it next to the preceding node.
	 * Alias of {@link #getColumn()}, but in inverted sankeys and org charts, the levels are laid out as rows.
	 * <p>
	 * Default: null
	 */
	public Number getLevel() {
		return getColumn();
	}

	/**
	 * Sets an optional level index of where to place the node.
	 * The default behaviour is to place it next to the preceding node.
	 * Alias of {@link #getColumn()}, but in inverted sankeys and org charts, the levels are laid out as rows.
	 */
	public void setLevel(Number level) {
		setColumn(level);
	}

	/**
	 * Returns the marker of this node, if any. Otherwise, create a new one.
	 */
	public Marker getMarker() {
		Marker marker = (Marker) getAttr(Attrs.marker);
		if (marker == null) {
			marker = new Marker();
			setMarker(marker);
		}
		return marker;
	}

	/**
	 * Sets the marker of this node.
	 */
	public void setMarker(Marker marker) {
		setAttr(Attrs.marker, marker);
	}

	/**
	 * Returns the mass of the node.
	 * By default, each node has mass equal to it's marker radius.
	 * Mass is used to determine how two connected nodes should affect each other:
	 *
	 * Attractive force is multiplied by the ratio of two connected nodes;
	 * if a big node has weights twice as the small one, then the small one will
	 * move towards the big one twice faster than the big one to the small one.
	 * <p>
	 * Default: null
	 */
	public Number getMass() {
		return getAttr(Attrs.mass, null).asNumber();
	}

	/**
	 * Sets the mass of the node.
	 * By default, each node has mass equal to it's marker radius.
	 * Mass is used to determine how two connected nodes should affect each other:
	 *
	 * Attractive force is multiplied by the ratio of two connected nodes;
	 * if a big node has weights twice as the small one, then the small one will
	 * move towards the big one twice faster than the big one to the small one.
	 */
	public void setMass(Number mass) {
		setAttr(Attrs.mass, mass);
	}

	/**
	 * Returns the name to display for the node in data labels and tooltips.
	 * Use this when the name is different from the id.
	 * Where the id must be unique for each node, this is not necessary for the name.
	 * <p>
	 * Default: null
	 */
	public String getName() {
		return getAttr(Attrs.name, null).asString();
	}

	/**
	 * Sets the name to display for the node in data labels and tooltips.
	 * Use this when the name is different from the id.
	 * Where the id must be unique for each node, this is not necessary for the name.
	 */
	public void setName(String name) {
		setAttr(Attrs.name, name);
	}

	/**
	 * In a horizontal layout, returns the vertical offset of a node in terms of weight.
	 * Positive values shift the node downwards, negative shift it upwards.
	 * In a vertical layout, like organization chart, the offset is horizontal.
	 *
	 * If a percentage string is given, the node is offset by the percentage of the node size plus {@code nodePadding}.
	 * <p>
	 * Default: 0
	 */
	public Object getOffset() {
		return getAttr(Attrs.offset, 0).asValue();
	}

	/**
	 * In a horizontal layout, sets the vertical offset of a node in terms of weight.
	 * Positive values shift the node downwards, negative shift it upwards.
	 * In a vertical layout, like organization chart, the offset is horizontal.
	 */
	public void setOffset(Number offset) {
		setAttr(Attrs.offset, offset);
	}

	/**
	 * In a horizontal layout, sets the vertical offset of a node in terms of weight.
	 * In a vertical layout, like organization chart, the offset is horizontal.
	 * The node is offset by the percentage of the node size plus {@code nodePadding}.
	 * <p>
	 * Default: 0
	 */
	public void setOffset(String offset) {
		setAttr(Attrs.offset, offset);
	}

	/**
	 * Returns the job title for the node card, will be inserted by the default dataLabel.nodeFormatter.
	 * <p>
	 * Default: null
	 */
	public String getTitle() {
		return getAttr(Attrs.title, null).asString();
	}

	/**
	 * Sets the job title for the node card, will be inserted by the default dataLabel.nodeFormatter.
	 */
	public void setTitle(String title) {
		setAttr(Attrs.title, title);
	}
}
