/** ResponseDataHandler.java.

	Purpose:
		
	Description:
		
	History:
		12:25:47 PM Jan 27, 2014, Created by jumperchen

Copyright (C) 2014 Potix Corporation. All Rights Reserved.
*/
package org.zkoss.chart.util;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.zkoss.chart.Axis;
import org.zkoss.chart.Charts;
import org.zkoss.chart.OptionDataEvent;
import org.zkoss.chart.OptionDataEvent.EventType;
import org.zkoss.chart.Optionable;
import org.zkoss.chart.Series;
import org.zkoss.json.JSONAware;
import org.zkoss.json.JavaScriptValue;

/**
 * A response data handler for {@link Charts}'s smart update changes.
 * The handler can merge the duplicated data changes and ignore the changes if
 * they are unnecessary. For example add a new series with multiple points that
 * the <code>addPoint</code> command changes are unnecessary in this case.
 * This handler is used when the component is going to output the data in JSON
 * string at rendering phase, so it is a good timing to merge the response data there.
 * 
 * @author jumperchen
 */
public class ResponseDataHandler implements JSONAware {
	private Map<Optionable, List<OptionDataEvent>> _queue = new LinkedHashMap<Optionable, List<OptionDataEvent>>();
	
	public void addQueue(OptionDataEvent event) {
		List<OptionDataEvent> list = _queue.get(event.getTarget());
		if (list == null) {
			list = new LinkedList<OptionDataEvent>();
			_queue.put(event.getTarget(), list);
		}
		// clear all response data when destroy an optionable object which is not initialized yet. 
		if (event.getType() == EventType.DESTROYED) {
			for(OptionDataEvent evt : list) {
				if (evt.getType() == EventType.INITIALIZED) {
					list.clear();
					return;
				}
			}
		}
		list.add(event);
	}

	public boolean isEmpty() {
		return _queue.isEmpty();
	}
	public void clear() {
		_queue.clear();
	}
	public String toJSONString() {
		if (isEmpty())
			return "null";
		boolean redrawChart = false;
		List<Object> response = new LinkedList<Object>();
		for (Map.Entry<Optionable, List<OptionDataEvent>> me : _queue.entrySet()) {
			boolean hasClassOperation = false;
			boolean hasUpdateFunction = false;
			if (me.getKey() instanceof Axis || me.getKey() instanceof Series) {// refer to PlotData.java
				for(OptionDataEvent evt : me.getValue()) {
					if (evt.getType() == EventType.INITIALIZED || evt.getType() == EventType.DESTROYED) {
						hasClassOperation = true;
						break;
					}
					if (evt.isJSUpldateCall()) {
						hasUpdateFunction = true;
					}
					if (!redrawChart && me.getKey() instanceof Series && evt.getType() == EventType.CHANGED && "visible".equals(evt.getKey())) {
						redrawChart = true;
					}
				}
			} else {
				for(OptionDataEvent evt : me.getValue()) {
					if (evt.getType() == EventType.INITIALIZED || evt.getType() == EventType.DESTROYED) {
						hasClassOperation = true;
						break;
					}
				}
			}
			
			if (hasClassOperation) {
				List<OptionDataEvent> value = me.getValue();
				for (Iterator<OptionDataEvent> it = value.iterator(); it.hasNext();) {
					OptionDataEvent evt = it.next();
					if (evt.getType() != EventType.INITIALIZED && evt.getType() != EventType.DESTROYED) {
						it.remove();
					}
				}
			}
			if (hasUpdateFunction) {
				Map<String, OptionDataEvent> commands = new HashMap<String, OptionDataEvent>();
				List<OptionDataEvent> value = me.getValue();
				for (Iterator<OptionDataEvent> it = value.iterator(); it.hasNext();) {
					OptionDataEvent evt = it.next();
					if (!evt.isJSUpldateCall()) {
						it.remove();
					} else if (me.getKey() instanceof Axis) {
						// merge duplicate command, like this.xAxis[0].update({"categories"...
						if (evt.getType() == EventType.CHANGED) {
							commands.put(evt.getKey(), evt);
						}
					}
				}
				// related to B-ZKCHARTS-10.zul test case, we don't need to generate too much xAxis[0].update commands
				if (!commands.isEmpty()) {
					for (Iterator<OptionDataEvent> it = value.iterator(); it.hasNext();) {
						OptionDataEvent evt = it.next();
						OptionDataEvent latestEvt = commands.get(evt.getKey());
						if (latestEvt != null && latestEvt != evt)
							it.remove();
					}
				}
			}
			response.addAll(me.getValue());
		}
		if (redrawChart) {
			response.add(new JavaScriptValue("function () {this.redraw();}"));
		}
		return OptionsList.toJSONString(response);
	}
}
